# -*- coding: utf-8 -*-
from accelbrainbase.observable_data import ObservableData
from accelbrainbase.iteratable_data import IteratableData
from accelbrainbase.regularizatable_data import RegularizatableData
from accelbrainbase.computable_loss import ComputableLoss
import numpy as np
from logging import getLogger
import torch
from torch import nn
from torch.optim.adam import Adam


class AttentionModel(nn.Module, ObservableData):
    '''
    Attention Model.

    References:
        - Bahdanau, D., Cho, K., & Bengio, Y. (2014). Neural machine translation by jointly learning to align and translate. arXiv preprint arXiv:1409.0473.
        - Floridi, L., & Chiriatti, M. (2020). GPT-3: Its nature, scope, limits, and consequences. Minds and Machines, 30(4), 681-694.
        - Miller, A., Fisch, A., Dodge, J., Karimi, A. H., Bordes, A., & Weston, J. (2016). Key-value memory networks for directly reading documents. arXiv preprint arXiv:1606.03126.
        - Radford, A., Narasimhan, K., Salimans, T., & Sutskever, I. (2018) Improving Language Understanding by Generative Pre-Training. OpenAI (URL: https://s3-us-west-2.amazonaws.com/openai-assets/research-covers/language-unsupervised/language_understanding_paper.pdf)
        - Radford, A., Wu, J., Child, R., Luan, D., Amodei, D., & Sutskever, I. (2019). Language models are unsupervised multitask learners. OpenAI blog, 1(8), 9.
        - Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., & Polosukhin, I. (2017). Attention is all you need. arXiv preprint arXiv:1706.03762.

    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    __depth_dim = None

    def get_depth_dim(self):
        ''' getter '''
        return self.__depth_dim
    
    def set_depth_dim(self, value):
        ''' setter '''
        self.__depth_dim = value

    depth_dim = property(get_depth_dim, set_depth_dim)

    def __init__(
        self,
        depth_dim,
        computable_loss,
        initializer_f=None,
        optimizer_f=None,
        dropout_rate=0.5,
        learning_rate=1e-05,
        ctx="cpu",
        regularizatable_data_list=[],
        not_init_flag=False,
        output_nn=None,
    ):
        '''
        Init.

        Args:
            depth_dim:                      `int` of dimension of dense layer.
            computable_loss:                is-a `ComputableLoss` or `gluon.loss.Loss`.
            dropout_rate:                   `float` of dropout rate.
            initializer:                    is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.

            optimizer_name:                 `str` of name of optimizer.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            regularizatable_data_list:           `list` of `RegularizatableData`.
            not_init_flag:                  `bool` of whether initialize parameters or not.
        '''
        if isinstance(depth_dim, int) is False:
            raise TypeError("The type of `depth_dim` must be `int`.")
        if isinstance(computable_loss, ComputableLoss) is False and isinstance(computable_loss, nn.modules.loss._Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `gluon.loss.Loss`.")

        super(AttentionModel, self).__init__()
        self.__computable_loss = computable_loss
        self.initializer_f = initializer_f
        self.optimizer_f = optimizer_f
        self.dropout = nn.Dropout(p=dropout_rate)
        self.softmax = nn.Softmax(dim=-1)
        self.output_nn = output_nn

        for v in regularizatable_data_list:
            if isinstance(v, RegularizatableData) is False:
                raise TypeError("The type of values of `regularizatable_data_list` must be `RegularizatableData`.")
        self.__regularizatable_data_list = regularizatable_data_list

        self.__ctx = ctx
        self.ctx = ctx
        self.__learning_rate = learning_rate

        self.depth_dim = depth_dim

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        self.__input_dim = None
        self.__not_init_flag = not_init_flag

        self.epoch = 0

    def initialize_params(self, input_dim):
        if self.__input_dim is not None:
            return

        self.__input_dim = input_dim

        self.query_dense_layer = nn.Linear(
            input_dim,
            self.depth_dim,
            bias=False,
        )
        if self.initializer_f is not None:
            self.query_dense_layer.weight = self.initializer_f(
                self.query_dense_layer.weight
            )
        else:
            self.query_dense_layer.weight = torch.nn.init.xavier_normal_(
                self.query_dense_layer.weight,
                gain=1.0
            )

        self.key_dense_layer = nn.Linear(
            input_dim,
            self.depth_dim,
            bias=False,
        )
        if self.initializer_f is not None:
            self.key_dense_layer.weight = self.initializer_f(
                self.key_dense_layer.weight
            )
        else:
            self.key_dense_layer.weight = torch.nn.init.xavier_normal_(
                self.key_dense_layer.weight,
                gain=1.0
            )

        self.value_dense_layer = nn.Linear(
            input_dim,
            self.depth_dim,
            bias=False,
        )
        if self.initializer_f is not None:
            self.value_dense_layer.weight = self.initializer_f(
                self.value_dense_layer.weight
            )
        else:
            self.value_dense_layer.weight = torch.nn.init.xavier_normal_(
                self.value_dense_layer.weight,
                gain=1.0
            )

        self.output_dense_layer = nn.Linear(
            self.depth_dim,
            self.depth_dim,
            bias=False,
        )
        if self.initializer_f is not None:
            self.output_dense_layer.weight = self.initializer_f(
                self.output_dense_layer.weight
            )
        else:
            self.output_dense_layer.weight = torch.nn.init.xavier_normal_(
                self.output_dense_layer.weight,
                gain=1.0
            )

        self.to(self.__ctx)

        if self.init_deferred_flag is False:
            if self.__not_init_flag is False:
                if self.optimizer_f is not None:
                    self.optimizer = self.optimizer_f(
                        self.parameters()
                    )
                else:
                    self.optimizer = Adam(
                        self.parameters(), 
                        lr=self.__learning_rate,
                    )

    def learn(self, iteratable_data):
        '''
        Learn samples drawn by `IteratableData.generate_learned_samples()`.

        Args:
            iteratable_data:     is-a `IteratableData`.
        '''
        if isinstance(iteratable_data, IteratableData) is False:
            raise TypeError("The type of `iteratable_data` must be `IteratableData`.")

        self.__loss_list = []
        learning_rate = self.__learning_rate

        pre_batch_observed_arr = None
        pre_test_batch_observed_arr = None
        try:
            epoch = self.epoch
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr in iteratable_data.generate_learned_samples():
                # Self-Attention.
                self.__batch_size = batch_observed_arr.shape[0]
                self.__seq_len = batch_observed_arr.shape[1]
                if len(batch_observed_arr.shape) == 2:
                    batch_observed_arr = torch.unsqueeze(batch_observed_arr, axis=1)
                elif len(batch_observed_arr.shape) > 3:
                    batch_observed_arr = batch_observed_arr.reshape((
                        self.__batch_size,
                        self.__seq_len,
                        -1
                    ))

                self.initialize_params(
                    input_dim=batch_observed_arr.shape[2]
                )
                self.epoch = epoch

                if self.output_nn is not None:
                    if hasattr(self.output_nn, "optimizer") is False:
                        _ = self.inference(batch_observed_arr, batch_observed_arr)

                self.optimizer.zero_grad()
                if self.output_nn is not None:
                    self.output_nn.optimizer.zero_grad()

                pred_arr = self.inference(batch_observed_arr, batch_observed_arr)
                loss = self.compute_loss(
                    pred_arr,
                    batch_target_arr
                )
                loss.backward()
                if self.output_nn is not None:
                    self.output_nn.optimizer.step()
                self.optimizer.step()
                self.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    with torch.inference_mode():
                        if len(test_batch_observed_arr.shape) == 2:
                            test_batch_observed_arr = torch.unsqueeze(test_batch_observed_arr, axis=1)

                        test_pred_arr = self.inference(test_batch_observed_arr, test_batch_observed_arr)

                        test_loss = self.compute_loss(
                            test_pred_arr,
                            test_batch_target_arr
                        )
                    _loss = loss.to('cpu').detach().numpy().copy()
                    _test_loss = test_loss.to('cpu').detach().numpy().copy()

                    self.__loss_list.append((_loss, _test_loss))
                    self.__logger.debug("Epochs: " + str(epoch + 1) + " Train loss: " + str(_loss) + " Test loss: " + str(_test_loss))
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.__logger.debug("end. ")
        self.epoch = epoch

    def inference(self, observed_arr, memory_arr):
        '''
        Inference samples drawn by `IteratableData.generate_inferenced_samples()`.

        Args:
            observed_arr:   Array like or sparse matrix as the observed data points.
            memory_arr:     Array like or sparse matrix as the observed data points.

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self(observed_arr, memory_arr)

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `mxnet.ndarray` or `mxnet.symbol`.
            labeled_arr:    `mxnet.ndarray` or `mxnet.symbol`.

        Returns:
            loss.
        '''
        return self.__computable_loss(pred_arr, labeled_arr)

    def regularize(self):
        '''
        Regularization.
        '''
        if len(self.__regularizatable_data_list) > 0:
            params_dict = self.extract_learned_dict()
            for regularizatable in self.__regularizatable_data_list:
                params_dict = regularizatable.regularize(params_dict)

            for k, params in params_dict.items():
                self.load_state_dict({k: params}, strict=False)

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_dict = {}
        for k in self.state_dict().keys():
            params_dict.setdefault(k, self.state_dict()[k])

        return params_dict

    def forward(self, x, m):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of observed data points.
            m:      `mxnet.ndarray` of observed data points. The shape is (batch_size, length of memory, depth).
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        # Self-Attention.
        self.__batch_size = x.shape[0]
        self.__seq_len = x.shape[1]
        if len(x.shape) == 2:
            x = torch.unsqueeze(x, axis=1)
        elif len(x.shape) > 3:
            x = x.reshape((
                self.__batch_size,
                self.__seq_len,
                -1
            ))

        m = m.reshape_as(x)

        self.initialize_params(
            input_dim=x.shape[2]
        )

        query_arr = self.query_dense_layer(x)
        key_arr = self.key_dense_layer(m)
        value_arr = self.value_dense_layer(m)
        logit_arr = torch.bmm(query_arr, key_arr.reshape((
            key_arr.shape[0],
            key_arr.shape[2],
            key_arr.shape[1]
        )))
        attention_weight_arr = self.softmax(logit_arr)
        attention_weight_arr = self.dropout(attention_weight_arr)
        attention_output_arr = torch.bmm(
            attention_weight_arr, 
            value_arr.reshape((
                value_arr.shape[0],
                value_arr.shape[2],
                value_arr.shape[1]
            ))
        )
        output_arr = self.output_dense_layer(attention_output_arr)

        if self.output_nn is not None:
            output_arr = self.output_nn(output_arr)

        return output_arr

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not. '''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not. '''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)



# -*- coding: utf-8 -*-
from accelbrainbase.samplabledata.policy_sampler import PolicySampler
from accelbrainbase.iteratabledata._mxnet.unlabeled_t_hot_txt_iterator import UnlabeledTHotTXTIterator

import mxnet.ndarray as nd
import mxnet as mx
import numpy as np


class LabeledSummarizationPolicy(PolicySampler):
    '''
    Policy sampler for the Deep Q-learning to evaluate the value of the 
    "action" of selecting the image with the highest similarity based on 
    the "state" of observing an image.

    The state-action value is proportional to the similarity between the previously 
    observed image and the currently selected image.

    This class calculates the image similarity by mean squared error of images.
    '''

    # is-a `UnlabeledTHotTXTIterator`.
    __unlabeled_t_hot_txt_iterator = None

    def get_unlabeled_t_hot_txt_iterator(self):
        ''' getter for `UnlabeledTHotTXTIterator`. '''
        if isinstance(self.__unlabeled_t_hot_txt_iterator, UnlabeledTHotTXTIterator) is False:
            raise TypeError("The type of `__unlabeled_t_hot_txt_iterator` must be `UnlabeledTHotTXTIterator`.")
        return self.__unlabeled_t_hot_txt_iterator
    
    def set_unlabeled_t_hot_txt_iterator(self, value):
        ''' setter for `UnlabeledTHotTXTIterator`.'''
        if isinstance(value, UnlabeledTHotTXTIterator) is False:
            raise TypeError("The type of `__unlabeled_t_hot_txt_iterator` must be `UnlabeledTHotTXTIterator`.")
        self.__unlabeled_t_hot_txt_iterator = value

    unlabeled_t_hot_txt_iterator = property(get_unlabeled_t_hot_txt_iterator, set_unlabeled_t_hot_txt_iterator)

    def __init__(
        self,
        txt_path_list,
        abstract_pos="top",
        s_a_dist_weight=0.3
    ):
        txt_list = []
        for txt_path in txt_path_list:
            with open(txt_path) as f:
                txt_list.append(f.read())
        self.__txt_list = txt_list
        self.__abstract_pos = abstract_pos
        self.__s_a_dist_weight = s_a_dist_weight

    def draw(self):
        '''
        Draw samples from distribtions.
        
        Returns:
            `Tuple` of `mx.nd.array`s.
        '''
        observed_arr, label_arr = None, None
        for arr_tuple in self.unlabeled_t_hot_txt_iterator.generate_learned_samples():
            _meta_data_arr = self.unlabeled_t_hot_txt_iterator.token_arr[
                arr_tuple[0].asnumpy().argmax(axis=-1)
            ]

            if observed_arr is None:
                observed_arr = nd.expand_dims(arr_tuple[0], axis=1)
                meta_data_arr = np.expand_dims(_meta_data_arr, axis=1)
            else:
                observed_arr = nd.concat(
                    observed_arr, 
                    nd.expand_dims(arr_tuple[0], axis=1), 
                    dim=1
                )
                meta_data_arr = np.concatenate(
                    [
                        meta_data_arr,
                        np.expand_dims(_meta_data_arr, axis=1)
                    ],
                    axis=1
                )

            if observed_arr.shape[1] >= self.possible_n:
                break

        return observed_arr, meta_data_arr

    def observe_state(self, state_arr, meta_data_arr):
        '''
        Observe states of agents in last epoch.

        Args:
            state_arr:      Tensor of state.
            meta_data_arr:  meta data of the state.
        '''
        self.__state_arr = state_arr
        self.__state_meta_data_arr = meta_data_arr

    def observe_reward_value(
        self, 
        state_arr, 
        action_arr,
        meta_data_arr=None,
    ):
        '''
        Compute the reward value.
        
        Args:
            state_arr:              Tensor of state.
            action_arr:             Tensor of action.
            meta_data_arr:          Meta data of actions.

        Returns:
            Reward value.
        '''
        if state_arr is not None:
            t_hot_loss = -nd.mean(
                nd.flatten(state_arr) * nd.flatten(action_arr),
                axis=0, 
                exclude=True
            )
            reward_value_arr = t_hot_loss
            reward_value_arr = nd.expand_dims(reward_value_arr, axis=1)
        else:
            reward_value_arr = nd.zeros((
                action_arr.shape[0],
                1
            ), ctx=action_arr.context)

        if meta_data_arr is not None:
            add_reward_arr = nd.zeros((action_arr.shape[0], 1), ctx=action_arr.context)
            for batch in range(meta_data_arr.shape[0]):
                keyword = "".join(meta_data_arr[batch].reshape(1, -1).tolist()[0])
                reward = 0.0
                for i in range(len(self.__txt_list)):
                    key = self.__txt_list[i].index(keyword)
                    reward = reward + ((len(self.__txt_list[i]) - key) / len(self.__txt_list[i]))
                    reward = reward + (self.__txt_list[i].count(keyword) / len(self.__txt_list[i]))
                add_reward_arr[batch] = reward / len(self.__txt_list)
        else:
            add_reward_arr = nd.zeros((
                meta_data_arr.shape[0], 1), 
                ctx=meta_data_arr.context
            )

        reward_value_arr = (reward_value_arr * self.__s_a_dist_weight) + (add_reward_arr * (1 - self.__s_a_dist_weight))
        reward_value_arr = nd.tanh(reward_value_arr)
        return reward_value_arr

    def update_state(
        self, 
        action_arr, 
        meta_data_arr=None
    ):
        '''
        Update state.
        
        This method can be overrided for concreate usecases.

        Args:
            action_arr:     action in `self.t`.
            meta_data_arr:  meta data of the action.
        
        Returns:
            Tuple data.
            - state in `self.t+1`.
            - meta data of the state.
        '''
        return action_arr, meta_data_arr

    def check_the_end_flag(self, state_arr, meta_data_arr=None):
        '''
        Check the end flag.
        
        If this return value is `True`, the learning is end.

        As a rule, the learning can not be stopped.
        This method should be overrided for concreate usecases.

        Args:
            state_arr:      state in `self.t`.
            meta_data_arr:  meta data of the state.

        Returns:
            bool
        '''
        # As a rule, the learning can not be stopped.
        return False



# -*- coding: utf-8 -*-
from accelbrainbase.controllable_model import ControllableModel

from accelbrainbase.observabledata._torch.adversarialmodel.generative_model import GenerativeModel
from accelbrainbase.observabledata._torch.adversarialmodel.discriminative_model import DiscriminativeModel
from accelbrainbase.samplabledata.true_sampler import TrueSampler

from accelbrainbase.computableloss._torch.generator_loss import GeneratorLoss
from accelbrainbase.computableloss._torch.discriminator_loss import DiscriminatorLoss

import numpy as np
from logging import getLogger

import torch
from torch import nn
from torch.optim.sgd import SGD


class GANController(nn.Module, ControllableModel):
    '''
    The Generative Adversarial Networks(GANs).

    The Generative Adversarial Networks(GANs) (Goodfellow et al., 2014) framework establishes 
    a min-max adversarial game between two neural networks – a generative model, `G`, and a 
    discriminative model, `D`. The discriminator model, `D(x)`, is a neural network that computes 
    the probability that a observed data point `x` in data space is a sample from the data 
    distribution (positive samples) that we are trying to model, rather than a sample from our 
    generative model (negative samples). 
    
    Concurrently, the generator uses a function `G(z)` that maps samples `z` from the prior `p(z)` to 
    the data space. `G(z)` is trained to maximally confuse the discriminator into believing that 
    samples it generates come from the data distribution. The generator is trained by leveraging 
    the gradient of `D(x)` w.r.t. x, and using that to modify its parameters.

    The Conditional GANs (or cGANs) is a simple extension of the basic GAN model which allows 
    the model to condition on external information. This makes it possible to engage the learned 
    generative model in different "modes" by providing it with different contextual 
    information (Gauthier, J. 2014).

    This model can be constructed by simply feeding the data, `y`, to condition on to both the generator 
    and discriminator. In an unconditioned generative model, because the maps samples `z` from the prior 
    `p(z)` are drawn from uniform or normal distribution, there is no control on modes of the data being 
    generated. On the other hand, it is possible to direct the data generation process by conditioning 
    the model on additional information (Mirza, M., & Osindero, S. 2014).

    References:
        - Gauthier, J. (2014). Conditional generative adversarial nets for convolutional face generation. Class Project for Stanford CS231N: Convolutional Neural Networks for Visual Recognition, Winter semester, 2014(5), 2.
        - Goodfellow, I., Pouget-Abadie, J., Mirza, M., Xu, B., Warde-Farley, D., Ozair, S., ... & Bengio, Y. (2014). Generative adversarial nets. In Advances in neural information processing systems (pp. 2672-2680).
        - Makhzani, A., Shlens, J., Jaitly, N., Goodfellow, I., & Frey, B. (2015). Adversarial autoencoders. arXiv preprint arXiv:1511.05644.
        - Mirza, M., & Osindero, S. (2014). Conditional generative adversarial nets. arXiv preprint arXiv:1411.1784.
        - Salimans, T., Goodfellow, I., Zaremba, W., Cheung, V., Radford, A., & Chen, X. (2016). Improved techniques for training gans. In Advances in neural information processing systems (pp. 2234-2242).
        - Zhao, J., Mathieu, M., & LeCun, Y. (2016). Energy-based generative adversarial network. arXiv preprint arXiv:1609.03126.
        - Warde-Farley, D., & Bengio, Y. (2016). Improving generative adversarial networks with denoising feature matching.

    '''
    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self,
        true_sampler,
        generative_model,
        discriminative_model,
        discriminator_loss,
        generator_loss=None,
        feature_matching_loss=None,
        discriminator_optimizer_f=None,
        generator_optimizer_f=None,
        learning_rate=1e-05,
        ctx="cpu",
        not_init_flag=False,
    ):
        '''
        Init.

        Args:
            true_sampler:                   is-a `TrueSampler`.
            generative_model:               is-a `GenerativeModel`.
            discriminative_model:           is-a `DiscriminativeModel`.
            generator_loss:                 is-a `GeneratorLoss`.
            discriminator_loss:             is-a `DiscriminatorLoss`.
            feature_matching_loss:          is-a `mxnet.gluon.loss`.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
                                            
            optimizer_name:                 `str` of name of optimizer.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.

        '''
        if isinstance(true_sampler, TrueSampler) is False:
            raise TypeError("The type of `true_sampler` must be `TrueSampler`.")
        if isinstance(generative_model, GenerativeModel) is False:
            raise TypeError("The type of `generative_model` must be `GenerativeModel`.")
        if isinstance(discriminative_model, DiscriminativeModel) is False:
            raise TypeError("The type of `discriminative_model` must be `DiscriminativeModel`.")

        if generator_loss is None:
            generator_loss = GeneratorLoss(weight=1.0)

        if isinstance(generator_loss, GeneratorLoss) is False:
            raise TypeError("The type of `generator_loss` must be `GeneratorLoss`.")
        if isinstance(discriminator_loss, DiscriminatorLoss) is False:
            raise TypeError("The type of `discriminator_loss` must be `DiscriminatorLoss`.")

        super(GANController, self).__init__()

        self.__true_sampler = true_sampler
        self.__generative_model = generative_model
        self.__discriminative_model = discriminative_model
        self.__generator_loss = generator_loss
        self.__discriminator_loss = discriminator_loss
        self.__feature_matching_loss = feature_matching_loss
        self.__learning_rate = learning_rate

        logger = getLogger("accelbrainbase")
        self.__logger = logger
        self.__not_init_flag = not_init_flag

        if self.init_deferred_flag is False:
            if self.__not_init_flag is False:
                if discriminator_optimizer_f is not None:
                    self.discriminator_optimizer = discriminator_optimizer_f(
                        self.__discriminative_model.parameters()
                    )
                else:
                    self.discriminator_optimizer = SGD(
                        self.__discriminative_model.parameters(),
                        lr=self.__learning_rate
                    )

                if generator_optimizer_f is not None:
                    self.generator_optimizer = generator_optimizer_f(
                        self.generative_model.parameters()
                    )
                else:
                    self.generator_optimizer = SGD(
                        self.generative_model.parameters(),
                        lr=self.__learning_rate
                    )

        self.__total_iter_n = 0
        self.__generative_loss_arr = None
        self.__discriminative_loss_arr = None
        self.__posterior_logs_arr = None
        self.__feature_matching_loss_arr = None

    def learn(
        self, 
        iter_n=1000,
        k_step=10,
    ):
        '''
        Learning.

        Args:
            iter_n:                         `int` of the number of training iterations.
            k_step:                         `int` of the number of learning of the `discriminative_model`.
        '''

        g_logs_list = []
        d_logs_list = []
        feature_matching_loss_list = []

        posterior_logs_list = []

        learning_rate = self.__learning_rate

        try:
            for n in range(iter_n):
                if (n + 1) % 100 == 0 or n < 100:
                    self.__logger.debug("-" * 100)
                    self.__logger.debug("Iterations: (" + str(n+1) + "/" + str(iter_n) + ")")
                    self.__logger.debug("-" * 100)
                    self.__logger.debug("The discriminator's turn.")
                    self.__logger.debug("-" * 100)

                loss, posterior = self.train_discriminator(k_step)
                if isinstance(loss, torch.Tensor) is True:
                    _loss = loss.to('cpu').detach().numpy()
                    _posterior = posterior.to('cpu').detach().numpy()
                else:
                    _loss = loss
                    _posterior = posterior

                d_logs_list.append(_loss)
                posterior_logs_list.append(_posterior)

                loss = self.train_by_feature_matching(k_step)
                if isinstance(loss, torch.Tensor) is True:
                    _loss = loss.to('cpu').detach().numpy()
                else:
                    _loss = loss
                feature_matching_loss_list.append(_loss)

                if (n + 1) % 100 == 0 or n < 100:
                    if len(posterior_logs_list) > 0:
                        self.__logger.debug("The discriminator's posterior(mean): " + str(posterior_logs_list[-1]))
                    if len(d_logs_list) > 0:
                        self.__logger.debug("The discriminator's loss(mean): " + str(d_logs_list[-1]))
                    if len(feature_matching_loss_list) > 0:
                        self.__logger.debug("The discriminator's feature matching loss(mean): " + str(feature_matching_loss_list[-1]))

                    self.__logger.debug("-" * 100)
                    self.__logger.debug("The generator's turn.")
                    self.__logger.debug("-" * 100)

                loss, posterior = self.train_generator()
                if isinstance(loss, torch.Tensor) is True:
                    _loss = loss.to('cpu').detach().numpy()
                    _posterior = posterior.to('cpu').detach().numpy()
                else:
                    _loss = loss
                    _posterior = posterior

                g_logs_list.append(_loss)
                posterior_logs_list.append(_posterior)

                if (n + 1) % 100 == 0 or n < 100:
                    if len(g_logs_list) > 0:
                        self.__logger.debug("The generator's loss(mean): " + str(g_logs_list[-1]))
                    if len(posterior_logs_list) > 0:
                        self.__logger.debug("The discriminator's posterior(mean): " + str(posterior_logs_list[-1]))

                self.__total_iter_n = self.__total_iter_n + 1

        except KeyboardInterrupt:
            print("Keyboard Interrupt.")

        if self.__generative_loss_arr is None:
            self.__generative_loss_arr = np.array(g_logs_list)
        else:
            self.__generative_loss_arr = np.r_[
                self.__generative_loss_arr,
                np.array(g_logs_list)
            ]

        if self.__discriminative_loss_arr is None:
            self.__discriminative_loss_arr = np.array(d_logs_list)
        else:
            self.__discriminative_loss_arr = np.r_[
                self.__discriminative_loss_arr,
                np.array(d_logs_list)
            ]

        if self.__posterior_logs_arr is None:
            self.__posterior_logs_arr = np.array(posterior_logs_list)
        else:
            self.__posterior_logs_arr = np.r_[
                self.__posterior_logs_arr,
                np.array(posterior_logs_list)
            ]

        if self.__feature_matching_loss_arr is None:
            self.__feature_matching_loss_arr = np.array(feature_matching_loss_list)
        else:
            self.__feature_matching_loss_arr = np.r_[
                self.__feature_matching_loss_arr,
                np.array(feature_matching_loss_list)
            ]

    def train_discriminator(
        self, 
        k_step, 
    ):
        '''
        Training for discriminator.

        Args:
            k_step:                         `int` of the number of learning of the `discriminative_model`.
        
        Returns:
            Tuple data.
            - discriminative loss.
            - discriminative posterior.
        '''
        d_loss = 0.0
        posterior = 0.0
        for k in range(k_step):
            true_arr = self.__true_sampler.draw()
            generated_arr = self.__generative_model.draw()

            self.discriminator_optimizer.zero_grad()

            true_posterior_arr = self.__discriminative_model.inference(true_arr)
            generated_posterior_arr = self.__discriminative_model.inference(generated_arr)
            loss = self.__discriminator_loss(
                true_posterior_arr,
                generated_posterior_arr
            )
            loss.backward()
            self.discriminator_optimizer.step()

            d_loss += loss
            posterior += torch.mean(generated_posterior_arr)

        d_loss = d_loss / k_step
        posterior = posterior / k_step
        return d_loss, posterior

    def train_by_feature_matching(self, k_step):
        feature_matching_loss = 0.0
        if self.feature_matching_loss is None:
            return feature_matching_loss

        for k in range(k_step):
            true_arr = self.__true_sampler.draw()
            generated_arr = self.__generative_model.draw()

            self.discriminator_optimizer.zero_grad()

            true_posterior_arr = self.__discriminative_model.inference(true_arr)
            generated_posterior_arr = self.__discriminative_model.inference(generated_arr)
            loss = self.feature_matching_loss(
                true_posterior_arr,
                generated_posterior_arr
            )
            loss.backward()
            self.discriminator_optimizer.step()

            feature_matching_loss += loss

        feature_matching_loss = feature_matching_loss / k_step
        return feature_matching_loss

    def train_generator(self):
        '''
        Train generator.
        
        Returns:
            Tuple data.
            - generative loss.
            - discriminative posterior.
        '''
        self.generator_optimizer.zero_grad()
        generated_arr = self.__generative_model.draw()
        generated_posterior_arr = self.__discriminative_model.inference(generated_arr)
        loss = self.__generator_loss(generated_posterior_arr)
        loss.backward()
        self.generator_optimizer.step()

        return loss, torch.mean(generated_posterior_arr)

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        torch.save(
            {
                'total_iter_n': self.__total_iter_n,
                'discriminative_model_state_dict': self.discriminative_model.state_dict(),
                'discriminator_optimizer_state_dict': self.discriminator_optimizer.state_dict(),
                'generative_model_state_dict': self.generative_model.state_dict(),
                'generator_optimizer_state_dict': self.generator_optimizer.state_dict(),
                'generative_loss': self.__generative_loss_arr,
                'discriminative_loss': self.__discriminative_loss_arr,
                'posterior_logs': self.__posterior_logs_arr,
                'feature_matching_loss': self.__feature_matching_loss_arr,
            }, 
            filename
        )

    def load_parameters(self, filename, ctx=None, strict=True):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            Context-manager that changes the selected device.
            strict:         Whether to strictly enforce that the keys in state_dict match the keys returned by this module’s state_dict() function. Default: `True`.
        '''
        checkpoint = torch.load(filename)
        self.__total_iter_n = checkpoint["total_iter_n"]
        self.discriminative_model.model.load_state_dict(
            checkpoint["discriminative_model_state_dict"], 
            strict=strict
        )
        self.generative_model.model.load_state_dict(
            checkpoint["generative_model_state_dict"], 
            strict=strict
        )

        self.discriminator_optimizer.load_state_dict(
            checkpoint['discriminator_optimizer_state_dict']
        )
        self.generator_optimizer.load_state_dict(
            checkpoint['generator_optimizer_state_dict']
        )

        self.__generative_loss_arr = checkpoint['generative_loss']
        self.__discriminative_loss_arr = checkpoint['discriminative_loss']
        self.__posterior_logs_arr = checkpoint['posterior_logs']
        self.__feature_matching_loss_arr = checkpoint['feature_matching_loss']

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_true_sampler(self):
        ''' getter for `TrueSampler`. '''
        return self.__true_sampler
    
    def set_true_sampler(self, value):
        ''' setter for `TrueSampler`.'''
        self.__true_sampler = value
    
    true_sampler = property(get_true_sampler, set_true_sampler)

    def get_generative_model(self):
        ''' getter for `GenerativeModel`. '''
        return self.__generative_model

    def set_generative_model(self, value):
        ''' getter for `GenerativeModel`. '''
        self.__generative_model = value

    generative_model = property(get_generative_model, set_generative_model)

    def get_discriminative_model(self):
        ''' getter for `DiscriminativeModel`. '''
        return self.__discriminative_model

    def set_discriminative_model(self, value):
        ''' getter for `DiscriminativeModel`. '''
        self.__discriminative_model = value

    discriminative_model = property(get_discriminative_model, set_discriminative_model)

    def get_feature_matching_loss(self):
        ''' getter for `FeatureMatchingLoss`. '''
        return self.__feature_matching_loss
    
    def set_feature_matching_loss(self, value):
        ''' setter for `FeatureMatchingLoss`. '''
        self.__feature_matching_loss = value
    
    feature_matching_loss = property(get_feature_matching_loss, set_feature_matching_loss)

    def get_generator_loss(self):
        ''' getter for `GeneratorLoss`. '''
        return self.__generator_loss

    def set_generator_loss(self, value):
        ''' setter for `GeneratorLoss`. '''
        self.__generator_loss = value

    generator_loss = property(get_generator_loss, set_generator_loss)

    def get_discriminator_loss(self):
        ''' getter for `DiscriminatorLoss`. '''
        return self.__discriminator_loss

    def set_discriminator_loss(self, value):
        ''' getter for `DiscriminatorLoss`. '''
        self.__discriminator_loss = value

    discriminator_loss = property(get_discriminator_loss, set_discriminator_loss)

    def get_generative_loss_arr(self):
        ''' getter for Generator's losses. '''
        return self.__generative_loss_arr
    
    generative_loss_arr = property(get_generative_loss_arr, set_readonly)

    def get_discriminative_loss_arr(self):
        ''' getter for Generator's losses.'''
        return self.__discriminative_loss_arr
    
    discriminative_loss_arr = property(get_discriminative_loss_arr, set_readonly)

    def get_posterior_logs_arr(self):
        ''' getter for logs of posteriors.'''
        return self.__posterior_logs_arr
    
    posterior_logs_arr = property(get_posterior_logs_arr, set_readonly)

    def get_feature_matching_loss_arr(self):
        ''' getter for logs of posteriors. '''
        return self.__feature_matching_loss_arr
    
    feature_matching_loss_arr = property(get_feature_matching_loss_arr, set_readonly)

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not.'''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not.'''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)



# -*- coding: utf-8 -*-
from mxnet.gluon.loss import Loss
import numpy as np
from mxnet.gluon.loss import _reshape_like, _apply_weighting
from accelbrainbase.computableloss._mxnet.discriminator_loss import DiscriminatorLoss


class EBDiscriminatorLoss(DiscriminatorLoss):
    '''
    Loss function of discriminators in Generative Adversarial Networks(GANs).
    '''

    def __init__(
        self, 
        weight=None,
        batch_axis=0, 
        margin=1.0,
        margin_decay_rate=0.1,
        margin_decay_epoch=50,
        **kwargs
    ):
        '''
        Init.

        Args:
            weight:                 Global scalar weight for loss.
            **kwargs:               Any parmaeter.
        '''
        super(EBDiscriminatorLoss, self).__init__(
            weight, 
            batch_axis, 
            **kwargs
        )
        self._weight = weight
        self.__margin = margin
        self.__margin_decay_rate = margin_decay_rate
        self.__margin_decay_epoch = margin_decay_epoch

        self.epoch = 0

    def compute(self, true_posterior_arr, generated_posterior_arr):
        '''
        Compute loss.

        Args:
            true_posterior_arr:            Real samples.
            generated_posterior_arr:       Generated samples.

        Returns:
            Tensor of losses.
        '''
        return self(true_posterior_arr, generated_posterior_arr)

    def hybrid_forward(
        self, 
        F, 
        true_posterior_arr,
        generated_posterior_arr,
        sample_weight=None
    ):
        '''
        Forward propagation, computing losses.

        Args:
            F:                          `mxnet.ndarray` or `mxnet.symbol`.
            true_posterior_arr:         `mxnet.ndarray` or `mxnet.symbol` of true posterior
                                        inferenced by the discriminator.

            generated_posterior_arr:    `mxnet.ndarray` or `mxnet.symbol` of fake posterior
                                        inferenced by the generator.

        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of loss.
        '''
        loss = true_posterior_arr + F.maximum(0, self.__margin - generated_posterior_arr)
        loss = _apply_weighting(F, loss, self._weight, sample_weight)

        self.epoch += 1
        if self.epoch % self.__margin_decay_epoch == 0:
            self.__margin = self.__margin * self.__margin_decay_rate

        return F.mean(loss, axis=self._batch_axis, exclude=True)



# -*- coding: utf-8 -*-
from abc import abstractmethod

import numpy as np
import torch

from logging import getLogger
import os
import random

from accelbrainbase.iteratable_data import IteratableData


class DatasetIterator(IteratableData):
    '''
    Iterator that draws from image files and generates `mxnet.ndarray` of unlabeled samples.
    '''

    __label_key = "labels"

    def __init__(
        self, 
        reserved_model,
        train_dataset,
        test_dataset,
        epochs=300,
        batch_size=20,
    ):
        '''
        Init.

        Args:
            image_extractor:                is-a `ImageExtractor`.
            dir_list:                       `list` of directories that store your image files in training.
                                            This class will not scan the directories recursively and consider that
                                            all image file will be sorted by any rule in relation to your sequencial modeling.

            test_dir_list:                  `list` of directories that store your image files in test.
                                            If `None`, this value will be equivalent to `dir_list`.
                                            This class will not scan the directories recursively and consider that
                                            all image file will be sorted by any rule in relation to your sequencial modeling.

            epochs:                         `int` of epochs of Mini-batch.
            batch_size:                     `int` of batch size of Mini-batch.

        '''
        if isinstance(train_dataset, torch.utils.data.Dataset) is False:
            raise TypeError()
        if isinstance(test_dataset, torch.utils.data.Dataset) is False:
            raise TypeError()

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        dataset_size = len(train_dataset)
        iter_n = int(epochs * max(dataset_size / batch_size, 1))

        forward_args_tuple = reserved_model.forward.__code__.co_varnames[
            :reserved_model.forward.__code__.co_argcount
        ]
        forward_args_list = [v for v in forward_args_tuple if v != "self"]
        if forward_args_list.count(self.__label_key) == 0:
            warnings.warn("`" + str(self.__label_key) + "` is not an argument in `reserved_model.forward`. This class was designed on the assumption that the `reserved_model` that calculate losses in `forward` method.")

        self.forward_args_list = forward_args_list
        self.train_dataset = train_dataset
        self.test_dataset = test_dataset
        self.iter_n = iter_n
        self.epochs = epochs
        self.batch_size = batch_size

    def convert_dataset_item_into_tuple(self, batch_dict):
        batch_list = [batch_dict[args] for args in self.forward_args_list]
        return tuple(batch_list)

    def generate_learned_samples(self):
        '''
        Draw and generate data.

        Returns:
            `Tuple` data. The shape is ...
            - `mxnet.ndarray` of observed data points in training.
            - `mxnet.ndarray` of supervised data in training.
            - `mxnet.ndarray` of observed data points in test.
            - `mxnet.ndarray` of supervised data in test.
        '''
        for _ in range(self.iter_n):
            training_batch_arr, test_batch_arr = None, None
            training_label_arr, test_label_arr = None, None
            training_batch_dict, test_batch_dict = {}, {}
            for batch_size in range(self.batch_size):
                train_i = np.random.randint(low=0, high=len(self.train_dataset))
                test_i = np.random.randint(low=0, high=len(self.test_dataset))

                training_data_dict = self.train_dataset[train_i]
                test_data_dict = self.test_dataset[test_i]
                
                if len(training_batch_dict) == 0:
                    for key, val in training_data_dict.items():
                        training_batch_dict.setdefault(
                            key, 
                            torch.unsqueeze(val, dim=0)
                        )
                else:
                    for key, val in training_data_dict.items():
                        training_batch_dict[key] = torch.cat(
                            (
                                training_batch_dict[key],
                                torch.unsqueeze(val, dim=0)
                            ),
                            dim=0
                        )

                if len(test_batch_dict) == 0:
                    for key, val in test_data_dict.items():
                        test_batch_dict.setdefault(
                            key, 
                            torch.unsqueeze(val, dim=0)
                        )
                else:
                    for key, val in test_data_dict.items():
                        test_batch_dict[key] = torch.cat(
                            (
                                test_batch_dict[key],
                                torch.unsqueeze(val, dim=0)
                            ),
                            dim=0
                        )

            training_batch_tuple = self.convert_dataset_item_into_tuple(
                training_batch_dict
            )
            test_batch_tuple = self.convert_dataset_item_into_tuple(
                test_batch_dict
            )
            yield training_batch_tuple, test_batch_tuple

    def generate_inferenced_samples(self):
        '''
        Draw and generate data.
        The targets will be drawn from all image file sorted in ascending order by file name.

        Returns:
            `Tuple` data. The shape is ...
            - `None`.
            - `None`.
            - `mxnet.ndarray` of observed data points in test.
            - file path.
        '''
        for test_i in range(len(self.test_dataset)):
            test_batch_arr = None
            test_label_arr = None
            test_batch_dict = {}
            for batch_size in range(self.batch_size):
                test_data_dict = self.test_dataset[test_i]

                if len(test_batch_dict) == 0:
                    for key, val in test_data_dict.items():
                        test_batch_dict.setdefault(
                            key, 
                            torch.unsqueeze(val, dim=0)
                        )
                else:
                    for key, val in test_data_dict.items():
                        test_batch_dict[key] = torch.cat(
                            (
                                test_batch_dict[key],
                                torch.unsqueeze(val, dim=0)
                            ),
                            dim=0
                        )

            test_batch_tuple = self.convert_dataset_item_into_tuple(
                test_batch_dict
            )
            yield test_batch_tuple



# -*- coding: utf-8 -*-
from accelbrainbase.controllable_model import ControllableModel
from accelbrainbase.observabledata._torch.function_approximator import FunctionApproximator
from accelbrainbase.samplabledata.policy_sampler import PolicySampler
from accelbrainbase.computable_loss import ComputableLoss

import numpy as np
from logging import getLogger
from abc import abstractmethod
import torch
from torch import nn


class DQLController(nn.Module, ControllableModel):
    '''
    Abstract base class to implement the Deep Q-Learning.

    The structure of Q-Learning is based on the Epsilon Greedy Q-Leanring algorithm,
    which is a typical off-policy algorithm.  In this paradigm, stochastic searching 
    and deterministic searching can coexist by hyperparameter `epsilon_greedy_rate` 
    that is probability that agent searches greedy. Greedy searching is deterministic 
    in the sensethat policy of agent follows the selection that maximizes the Q-Value.

    References:
        - https://code.accel-brain.com/Reinforcement-Learning/README.html#deep-q-network
        - Egorov, M. (2016). Multi-agent deep reinforcement learning.(URL: https://pdfs.semanticscholar.org/dd98/9d94613f439c05725bad958929357e365084.pdf)
        - Gupta, J. K., Egorov, M., & Kochenderfer, M. (2017, May). Cooperative multi-agent control using deep reinforcement learning. In International Conference on Autonomous Agents and Multiagent Systems (pp. 66-83). Springer, Cham.
        - Mnih, V., Kavukcuoglu, K., Silver, D., Graves, A., Antonoglou, I., Wierstra, D., & Riedmiller, M. (2013). Playing atari with deep reinforcement learning. arXiv preprint arXiv:1312.5602.

    '''

    def __init__(
        self,
        function_approximator,
        policy_sampler,
        computable_loss,
        ctx="cpu",
        recursive_learning_flag=False,
    ):
        '''
        Init.

        Args:
            function_approximator:          is-a `FunctionApproximator`.
            policy_sampler:                 is-a `PolicySampler`.
            computable_loss:                is-a `ComputableLoss` or `mxnet.gluon.loss`.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
                                            
            optimizer_name:                 `str` of name of optimizer.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.

        '''
        if isinstance(function_approximator, FunctionApproximator) is False:
            raise TypeError("The type of `function_approximator` must be `FunctionApproximator`.")
        if isinstance(policy_sampler, PolicySampler) is False:
            raise TypeError("The type of `policy_sampler` must be `PolicySampler`.")
        if isinstance(computable_loss, ComputableLoss) is False and isinstance(computable_loss, nn.modules.loss._Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `gluon.loss.Loss`.")

        super(DQLController, self).__init__()

        self.__function_approximator = function_approximator
        self.function_approximator = self.__function_approximator
        self.__policy_sampler = policy_sampler
        self.__computable_loss = computable_loss

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        self.__recursive_learning_flag = recursive_learning_flag
        self.__q_logs_arr = np.array([])

        self.t = 0
        self.epoch = 0

    def learn(
        self, 
        iter_n=100,
    ):
        '''
        Learning.

        Args:
            iter_n:     `int` of the number of training iterations.
        '''
        q_logs_list = []
        r_logs_list = []
        posterior_logs_list = []

        state_arr = None
        state_meta_data_arr = None

        try:
            for n in range(iter_n):
                if (n + 1) % 100 == 0 or n < 100:
                    self.__logger.debug("-" * 100)
                    self.__logger.debug("Iter: " + str(n + 1))

                # Draw samples of next possible actions from any distribution.
                # (batch, possible_n, dim1, dim2, ...)
                possible_action_arr, action_meta_data_arr = self.policy_sampler.draw()

                possible_reward_value_arr = None
                next_q_arr = None
                possible_predicted_q_arr = None

                for possible_i in range(possible_action_arr.shape[1]):
                    if action_meta_data_arr is not None:
                        meta_data_arr = action_meta_data_arr[:, possible_i]
                    else:
                        meta_data_arr = None

                    # Inference Q-Values.
                    _predicted_q_arr = self.function_approximator(
                        possible_action_arr[:, possible_i]
                    )
                    if possible_predicted_q_arr is None:
                        possible_predicted_q_arr = torch.unsqueeze(_predicted_q_arr, dim=1)
                    else:
                        possible_predicted_q_arr = torch.cat(
                            (
                                possible_predicted_q_arr,
                                torch.unsqueeze(_predicted_q_arr, dim=1),
                            ),
                            dim=1
                        )

                    # Observe reward values.
                    _reward_value_arr = self.policy_sampler.observe_reward_value(
                        state_arr, 
                        possible_action_arr[:, possible_i],
                        meta_data_arr=meta_data_arr,
                    )
                    if possible_reward_value_arr is None:
                        possible_reward_value_arr = torch.unsqueeze(_reward_value_arr, axis=1)
                    else:
                        possible_reward_value_arr = torch.cat(
                            (
                                possible_reward_value_arr,
                                torch.unsqueeze(_reward_value_arr, dim=1),
                            ),
                            dim=1
                        )

                    # Inference the Max-Q-Value in next action time.
                    self.policy_sampler.observe_state(
                        state_arr=possible_action_arr[:, possible_i],
                        meta_data_arr=meta_data_arr
                    )
                    next_possible_action_arr, _ = self.policy_sampler.draw()
                    next_next_q_arr = None

                    for possible_j in range(next_possible_action_arr.shape[1]):
                        with torch.inference_mode():
                            _next_next_q_arr = self.function_approximator(
                                next_possible_action_arr[:, possible_j]
                            )
                        if next_next_q_arr is None:
                            next_next_q_arr = torch.unsqueeze(
                                _next_next_q_arr,
                                dim=1
                            )
                        else:
                            next_next_q_arr = torch.cat(
                                (
                                    next_next_q_arr,
                                    torch.unsqueeze(
                                        _next_next_q_arr, 
                                        dim=1
                                    ),
                                ),
                                dim=1
                            )

                    next_max_q_arr = next_next_q_arr.max(dim=1, keepdim=True).values

                    if next_q_arr is None:
                        next_q_arr = torch.unsqueeze(
                            next_max_q_arr,
                            dim=1,
                        )
                    else:
                        next_q_arr = torch.cat(
                            (
                                next_q_arr,
                                torch.unsqueeze(
                                    next_max_q_arr,
                                    dim=1,
                                ),
                            ),
                            dim=1
                        )

                # Select action.
                selected_tuple = self.select_action(
                    possible_action_arr, 
                    possible_predicted_q_arr,
                    possible_reward_value_arr,
                    next_q_arr,
                    possible_meta_data_arr=action_meta_data_arr
                )
                action_arr, predicted_q_arr, reward_value_arr, next_q_arr, action_meta_data_arr = selected_tuple

                self.function_approximator.model.optimizer.zero_grad()

                predicted_q_arr = self.function_approximator(
                    action_arr
                )
                # Update real Q-Values.
                real_q_arr = self.update_q(
                    reward_value_arr,
                    next_q_arr
                )

                if self.__q_logs_arr.shape[0] > 0:
                    self.__q_logs_arr = np.r_[
                        self.__q_logs_arr,
                        np.array([
                            predicted_q_arr.to('cpu').detach().numpy().copy().mean(), 
                            real_q_arr.to('cpu').detach().numpy().copy().mean(),
                        ]).reshape(1, 2)
                    ]
                else:
                    self.__q_logs_arr = np.array([
                        predicted_q_arr.to('cpu').detach().numpy().copy().mean(), 
                        real_q_arr.to('cpu').detach().numpy().copy().mean()
                    ]).reshape(1, 2)

                # Learn Q-Values.
                loss = self.computable_loss(
                    predicted_q_arr, 
                    real_q_arr
                )
                if torch.isnan(loss).to('cpu').detach().numpy().copy().astype(int).sum() == 0:
                    loss.backward()
                    self.function_approximator.model.optimizer.step()
                    self.function_approximator.model.regularize()

                    if (n + 1) % 100 == 0 or n < 100:
                        self.__logger.debug("Reward value(mean): " + str(reward_value_arr.mean().to('cpu').detach().numpy().copy()))
                        self.__logger.debug("Predicted Q-value(mean): " + str(predicted_q_arr.mean().to('cpu').detach().numpy().copy()))
                        self.__logger.debug("Real Q-value(mean): " + str(real_q_arr.mean().to('cpu').detach().numpy().copy()))
                        self.__logger.debug("Loss of Q-Value(mean): " + str(loss.mean().to('cpu').detach().numpy().copy()))

                    if self.__recursive_learning_flag is True:
                        # Update State.
                        state_arr, state_meta_data_arr = self.policy_sampler.update_state(
                            action_arr, 
                            meta_data_arr=action_meta_data_arr
                        )
                        self.policy_sampler.observe_state(
                            state_arr=state_arr,
                            meta_data_arr=state_meta_data_arr
                        )
                else:
                    self.__logger.debug("The parameter update was skipped because the vanishing gradient problem or gradient explosion occurred.")

                # Epsode.
                self.t += 1

                if self.__recursive_learning_flag is True:
                    # Check.
                    end_flag = self.policy_sampler.check_the_end_flag(
                        state_arr, 
                        meta_data_arr=meta_data_arr
                    )
                    if end_flag is True:
                        break

        except KeyboardInterrupt:
            print("Keyboard Interrupt.")

    def inference(self, iter_n=100):
        '''
        Inference.

        Args:
            iter_n:     `int` of the number of training iterations.

        Returns:
            `list` of logs of states.
        '''
        state_arr_list = []
        q_value_arr_list = []
        state_arr = None
        try:
            for n in range(iter_n):
                # Draw samples of next possible actions from any distribution.
                # (batch, possible_n, dim1, dim2, ...)
                possible_action_arr, action_meta_data_arr = self.policy_sampler.draw()

                possible_reward_value_arr = None
                next_q_arr = None
                possible_predicted_q_arr = None

                for possible_i in range(possible_action_arr.shape[1]):
                    if action_meta_data_arr is not None:
                        meta_data_arr = action_meta_data_arr[:, possible_i]
                    else:
                        meta_data_arr = None

                    # Inference Q-Values.
                    _predicted_q_arr = self.function_approximator(
                        possible_action_arr[:, possible_i]
                    )
                    if possible_predicted_q_arr is None:
                        possible_predicted_q_arr = torch.unsqueeze(_predicted_q_arr, dim=1)
                    else:
                        possible_predicted_q_arr = torch.cat(
                            (
                                possible_predicted_q_arr,
                                torch.unsqueeze(_predicted_q_arr, dim=1),
                            ),
                            dim=1
                        )

                    # Observe reward values.
                    _reward_value_arr = self.policy_sampler.observe_reward_value(
                        state_arr, 
                        possible_action_arr[:, possible_i],
                        meta_data_arr=meta_data_arr,
                    )
                    if possible_reward_value_arr is None:
                        possible_reward_value_arr = torch.unsqueeze(_reward_value_arr, dim=1)
                    else:
                        possible_reward_value_arr = torch.cat(
                            (
                                possible_reward_value_arr,
                                torch.unsqueeze(_reward_value_arr, dim=1),
                            ),
                            dim=1
                        )

                    # Inference the Max-Q-Value in next action time.
                    self.policy_sampler.observe_state(
                        state_arr=possible_action_arr[:, possible_i],
                        meta_data_arr=meta_data_arr
                    )
                    next_possible_action_arr, _ = self.policy_sampler.draw()
                    next_next_q_arr = None

                    for possible_j in range(next_possible_action_arr.shape[1]):
                        _next_next_q_arr = self.function_approximator(
                            next_possible_action_arr[:, possible_j]
                        )
                        if next_next_q_arr is None:
                            next_next_q_arr = torch.unsqueeze(
                                _next_next_q_arr,
                                dim=1
                            )
                        else:
                            next_next_q_arr = torch.cat(
                                (
                                    next_next_q_arr,
                                    torch.unsqueeze(
                                        _next_next_q_arr, 
                                        dim=1
                                    ),
                                ),
                                dim=1
                            )

                    next_max_q_arr = next_next_q_arr.max(dim=1, keepdim=True).values

                    if next_q_arr is None:
                        next_q_arr = torch.unsqueeze(
                            next_max_q_arr,
                            dim=1
                        )
                    else:
                        next_q_arr = torch.cat(
                            (
                                next_q_arr,
                                torch.unsqueeze(
                                    next_max_q_arr,
                                    dim=1
                                ),
                            ),
                            dim=1
                        )

                # Select action.
                selected_tuple = self.select_action(
                    possible_action_arr, 
                    possible_predicted_q_arr,
                    possible_reward_value_arr,
                    next_q_arr,
                    possible_meta_data_arr=action_meta_data_arr
                )
                action_arr, predicted_q_arr, reward_value_arr, next_q_arr, action_meta_data_arr = selected_tuple

                # Update State.
                state_arr, state_meta_data_arr = self.policy_sampler.update_state(
                    action_arr, 
                    meta_data_arr=action_meta_data_arr
                )
                self.policy_sampler.observe_state(
                    state_arr=state_arr,
                    meta_data_arr=state_meta_data_arr
                )
                state_arr_list.append(state_arr)
                q_value_arr_list.append(predicted_q_arr)

                # Epsode.
                self.t += 1

                # Check.
                end_flag = self.policy_sampler.check_the_end_flag(
                    state_arr, 
                    meta_data_arr=meta_data_arr
                )
                if end_flag is True:
                    break

        except KeyboardInterrupt:
            print("Keyboard Interrupt.")

        return state_arr_list, q_value_arr_list

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_dict = {}
        for k in self.function_approximator.model.state_dict().keys():
            params_dict.setdefault(k, self.function_approximator.model.state_dict()[k])

        return params_dict

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        self.function_approximator.model.save_parameters(filename)

    def load_parameters(self, filename, ctx=None, strict=True):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            `mx.cpu()` or `mx.gpu()`.
            allow_missing:  `bool` of whether to silently skip loading parameters not represents in the file.
            ignore_extra:   `bool` of whether to silently ignre parameters from the file that are not present in this `Block`.
        '''
        self.function_approximator.model.load_parameters(filename, ctx=ctx, strict=strict)

    @abstractmethod
    def select_action(
        self, 
        possible_action_arr, 
        possible_predicted_q_arr, 
        possible_reward_value_arr,
        possible_meta_data_arr=None
    ):
        '''
        Select action by Q(state, action).

        Args:
            possible_action_arr:        Tensor of actions.
            possible_predicted_q_arr:             Tensor of Q-Values.
            possible_reward_value_arr:       Tensor of reward values.
            possible_meta_data_arr:          Meta data of the actions.

        Retruns:
            Tuple(`np.ndarray` of action., Q-Value)
        '''
        raise NotImplementedError("This method must be implemented.")

    def update_q(self, reward_value_arr, next_max_q_arr):
        '''
        Update Q.
        
        Args:
            reward_value_arr:   `np.ndarray` of reward values.
            next_max_q_arr:     `np.ndarray` of maximum Q-Values in next time step.
        
        Returns:
            `np.ndarray` of real Q-Values.
        '''
        reward_value_arr = reward_value_arr.reshape((reward_value_arr.shape[0], 1))
        next_max_q_arr = next_max_q_arr.reshape((next_max_q_arr.shape[0], 1))
        # Update Q-Value.
        return self.alpha_value * (reward_value_arr + (self.gamma_value * next_max_q_arr))

    __alpha_value = 1.0

    def get_alpha_value(self):
        '''
        getter
        Learning rate.
        '''
        if isinstance(self.__alpha_value, float) is False:
            raise TypeError("The type of __alpha_value must be float.")
        return self.__alpha_value

    def set_alpha_value(self, value):
        '''
        setter
        Learning rate.
        '''
        if isinstance(value, float) is False:
            raise TypeError("The type of __alpha_value must be float.")
        self.__alpha_value = value

    alpha_value = property(get_alpha_value, set_alpha_value)

    # Gamma Value.
    __gamma_value = 0.5

    def get_gamma_value(self):
        '''
        getter
        Gamma value.
        '''
        if isinstance(self.__gamma_value, float) is False:
            raise TypeError("The type of __gamma_value must be float.")
        return self.__gamma_value

    def set_gamma_value(self, value):
        '''
        setter
        Gamma value.
        '''
        if isinstance(value, float) is False:
            raise TypeError("The type of __gamma_value must be float.")
        self.__gamma_value = value

    gamma_value = property(get_gamma_value, set_gamma_value)

    __q_logs_arr = None

    def get_q_logs_arr(self):
        ''' getter '''
        return self.__q_logs_arr
    
    def set_q_logs_arr(self, values):
        ''' setter '''
        raise TypeError("The `q_logs_arr` must be read-only.")
    
    q_logs_arr = property(get_q_logs_arr, set_q_logs_arr)

    # is-a `PolicySampler`.
    __policy_sampler = None

    def get_policy_sampler(self):
        ''' getter for `PolicySampler` '''
        return self.__policy_sampler

    def set_policy_sampler(self, value):
        ''' setter for `PolicySampler` '''
        if isinstance(value, PolicySampler) is False:
            raise TypeError()
        self.__policy_sampler = value

    policy_sampler = property(get_policy_sampler, set_policy_sampler)

    def get_computable_loss(self):
        ''' getter for `ComputableLoss`. '''
        return self.__computable_loss
    
    def set_computable_loss(self, value):
        ''' setter for `ComputableLoss`. '''
        if isinstance(value, ComputableLoss) is False and isinstance(value, gluon.loss.Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `gluon.loss.Loss`.")
        self.__computable_loss = value

    computable_loss = property(get_computable_loss, set_computable_loss)



# -*- coding: utf-8 -*-
from accelbrainbase._mxnet._exception.init_deferred_error import InitDeferredError
from accelbrainbase.observabledata._mxnet.adversarial_model import AdversarialModel
from accelbrainbase.samplabledata.condition_sampler import ConditionSampler
from accelbrainbase.samplabledata.noise_sampler import NoiseSampler

from mxnet.gluon.block import HybridBlock
from mxnet import gluon
from mxnet import autograd
import numpy as np
import mxnet as mx
import mxnet.ndarray as nd
from mxnet import MXNetError
from logging import getLogger


class GenerativeModel(HybridBlock, AdversarialModel):
    '''
    Generative model, which draws samples from the fake distribution, 
    in the Generative Adversarial Networks(GANs).

    The Generative Adversarial Networks(GANs) (Goodfellow et al., 2014) framework establishes 
    a min-max adversarial game between two neural networks – a generative model, G, and a 
    discriminative model, D. The discriminator model, D(x), is a neural network that computes 
    the probability that a observed data point x in data space is a sample from the data 
    distribution (positive samples) that we are trying to model, rather than a sample from our 
    generative model (negative samples). 
    
    Concurrently, the generator uses a function G(z) that maps samples z from the prior p(z) to 
    the data space. G(z) is trained to maximally confuse the discriminator into believing that 
    samples it generates come from the data distribution. The generator is trained by leveraging 
    the gradient of D(x) w.r.t. x, and using that to modify its parameters.

    References:
        - Gauthier, J. (2014). Conditional generative adversarial nets for convolutional face generation. Class Project for Stanford CS231N: Convolutional Neural Networks for Visual Recognition, Winter semester, 2014(5), 2.
        - Goodfellow, I., Pouget-Abadie, J., Mirza, M., Xu, B., Warde-Farley, D., Ozair, S., ... & Bengio, Y. (2014). Generative adversarial nets. In Advances in neural information processing systems (pp. 2672-2680).
        - Makhzani, A., Shlens, J., Jaitly, N., Goodfellow, I., & Frey, B. (2015). Adversarial autoencoders. arXiv preprint arXiv:1511.05644.
        - Mirza, M., & Osindero, S. (2014). Conditional generative adversarial nets. arXiv preprint arXiv:1411.1784.
        - Salimans, T., Goodfellow, I., Zaremba, W., Cheung, V., Radford, A., & Chen, X. (2016). Improved techniques for training gans. In Advances in neural information processing systems (pp. 2234-2242).
        - Zhao, J., Mathieu, M., & LeCun, Y. (2016). Energy-based generative adversarial network. arXiv preprint arXiv:1609.03126.
        - Warde-Farley, D., & Bengio, Y. (2016). Improving generative adversarial networks with denoising feature matching.

    '''

    def __init__(
        self, 
        noise_sampler, 
        model, 
        initializer=None,
        condition_sampler=None,
        conditonal_dim=1,
        learning_rate=1e-05,
        optimizer_name="SGD",
        hybridize_flag=True,
        scale=1.0, 
        ctx=mx.gpu(), 
        **kwargs
    ):
        '''
        Init.

        Args:
            noise_sampler:                  is-a `NoiseSampler`.
            model:                          is-a `mxnet.gluon.hybrid.hybridblock.HybridBlock`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            condition_sampler:              is-a `ConditionSampler` of sampler to draw conditons from user-defined distributions.
            conditonal_dim:                 `int` of the dimension to be concated conditions and observed data points.
            learning_rate:                  `float` of learning rate.
            optimizer_name:                 `str` of name of optimizer.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            ctx:                            `mx.cpu()` or `mx.gpu()`.

        '''
        super(GenerativeModel, self).__init__()

        if initializer is None:
            self.initializer = mx.initializer.Xavier(
                rnd_type="gaussian", 
                factor_type="in", 
                magnitude=2
            )
        else:
            if isinstance(initializer, mx.initializer.Initializer) is False:
                raise TypeError("The type of `initializer` must be `mxnet.initializer.Initializer`.")
            self.initializer = initializer

        self.noise_sampler = noise_sampler
        self.model = model
        logger = getLogger("accelbrainbase")
        self.__logger = logger

        if self.init_deferred_flag is False:
            try:
                self.collect_params().initialize(self.initializer, force_reinit=True, ctx=ctx)
                self.trainer = gluon.Trainer(
                    self.collect_params(),
                    optimizer_name,
                    {
                        "learning_rate": learning_rate
                    }
                )
                if hybridize_flag is True:
                    self.model.hybridize()
                    if condition_sampler is not None:
                        if condition_sampler.model is not None:
                            condition_sampler.model.hybridize()

            except InitDeferredError:
                self.__logger.debug("The initialization should be deferred.")

        self.condition_sampler = condition_sampler
        self.conditonal_dim = conditonal_dim

    def collect_params(self, select=None):
        '''
        Overrided `collect_params` in `mxnet.gluon.HybridBlok`.
        '''
        params_dict = self.model.collect_params(select)
        return params_dict

    def draw(self):
        '''
        Draw samples from the fake distribution.

        Returns:
            `Tuple` of `mxnet.ndarray`s.
        '''
        if self.condition_sampler is None:
            return self.model(self.noise_sampler.draw())
        else:
            condition_arr, sampled_arr = self.condition_sampler.draw()
            if sampled_arr is not None:
                if self.noise_sampler is not None:
                    sampled_arr = sampled_arr + self.noise_sampler.draw()
                inferenced_arr = self.model(sampled_arr)

                generated_arr = nd.concat(
                    inferenced_arr,
                    condition_arr,
                    dim=self.conditonal_dim
                )
            else:
                if self.noise_sampler is not None:
                    condition_arr = condition_arr + self.noise_sampler.draw()
                generated_arr = self.model(condition_arr)

            return generated_arr

    # is-a `NoiseSampler`.
    __noise_sampler = None

    def get_noise_sampler(self):
        ''' getter '''
        return self.__noise_sampler

    def set_noise_sampler(self, value):
        ''' setter '''
        if isinstance(value, NoiseSampler) is False and value is not None:
            raise TypeError("The type of `noise_sampler` must be `NoiseSampler`.")
        self.__noise_sampler = value
    
    noise_sampler = property(get_noise_sampler, set_noise_sampler)

    # is-a `mxnet.gluon.hybrid.hybridblock.HybridBlock`.
    __model = None

    def get_model(self):
        ''' getter for `mxnet.gluon.hybrid.hybridblock.HybridBlock`.'''
        return self.__model
    
    def set_model(self, value):
        ''' setter for `mxnet.gluon.hybrid.hybridblock.HybridBlock`.'''
        self.__model = value
    
    model = property(get_model, set_model)

    # is-a `ConditionSampler` of sampler to draw conditons from user-defined distributions.
    __conditon_sampler = None

    def get_condition_sampler(self):
        ''' getter for `ConditionSampler` of sampler to draw conditons from user-defined distributions.'''
        return self.__conditon_sampler
    
    def set_condition_sampler(self, value):
        ''' setter for `ConditionSampler` of sampler to draw conditons from user-defined distributions.'''
        if isinstance(value, ConditionSampler) is False:
            raise TypeError("The type of `condition_sampler` must be `ConditionSampler`.")
        self.__conditon_sampler = value

    conditon_sampler = property(get_condition_sampler, set_condition_sampler)



# -*- coding: utf-8 -*-
from accelbrainbase.extractable_data import ExtractableData
import pandas as pd


class LabeledCSVExtractor(ExtractableData):
    '''
    CSV Extractor for labeled samples.
    '''

    # `str` of column of label
    __label_column = "label"

    def __init__(
        self, 
        label_column="label"
    ):
        '''
        Init.

        Args:
            label_column:        `str` of column of label.
        '''
        self.label_column = label_column

    def extract(
        self,
        path,
    ):
        '''
        Extract CSV file data.

        Args:
            path:     `str` of CSV files.
        
        Returns:
            Tuple data.
            - Observed data points.
            - label data.
        '''
        df = pd.read_csv(path)
        label_arr = df[self.label_column].values
        observed_arr = df[[col for col in df.columns if col != self.label_column]].values

        return observed_arr, label_arr

    def get_label_column(self):
        ''' getter of `str` of column of label. '''
        return self.__label_column

    def set_label_column(self, value):
        ''' setter of `str` of column of label. '''
        if isinstance(value, str):
            self.__label_column = value
        else:
            raise TypeError("The type of `label_column` must be `str`.")
    
    label_column = property(get_label_column, set_label_column)



# -*- coding: utf-8 -*-
from accelbrainbase.extractable_data import ExtractableData
import pandas as pd


class UnlabeledCSVExtractor(ExtractableData):
    '''
    CSV Extractor for unlabeled samples.
    '''

    def extract(
        self,
        path,
    ):
        '''
        Extract CSV file data.

        Args:
            path:     `str` of CSV files.
        
        Returns:
            Tuple data.
            - Observed data points.
            - label data.
        '''
        df = pd.read_csv(path)
        return df.values



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

def read_readme(file_name):
    from os import path
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, file_name), encoding='utf-8') as f:
        long_description = f.read()

    return long_description


setup(
    name='accel-brain-base',
    version='1.0.8',
    description='accel-brain-base is a basic library of the Deep Learning for rapid development at low cost. This library makes it possible to design and implement deep learning, which must be configured as a complex system, by combining a plurality of functionally differentiated modules such as a Deep Boltzmann Machines(DBMs), an Auto-Encoder, an Encoder/Decoder based on Long Short-Term Memory(LSTM), a Convolutional Auto-Encoder(CAE), and Transformer.',
    long_description=read_readme("README.md"),
    long_description_content_type='text/markdown',
    url='https://github.com/accel-brain/accel-brain-code/tree/master/Accel-Brain-Base/',
    author='accel-brain',
    author_email='info@accel-brain.co.jp',
    license='GPL2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Framework :: Robot Framework',
        'Topic :: Text Processing',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 3',
    ],
    keywords='Generative Adversarial Networks Adversarial Auto-Encoders autoencoder auto-encoder convolution deconvolution encoder decoder LSTM Deep Q-Learning Deep Reinforcement Learning Deep Boltzmann Machines Restricted Boltzmann Machine Ladder Networks semi-supervised learning DRCN Self-supervised attention transformer',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=['numpy', 'pandas'],
)



# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._torch.lstm_networks import LSTMNetworks
from accelbrainbase.iteratable_data import IteratableData
from accelbrainbase.regularizatable_data import RegularizatableData
from accelbrainbase.computable_loss import ComputableLoss
import numpy as np
import torch
from torch import nn
from logging import getLogger


class EncoderDecoder(LSTMNetworks):
    '''
    Encoder/Decoder based on LSTM networks.
    
    This library provides Encoder/Decoder based on LSTM, 
    which is a reconstruction model and makes it possible to extract 
    series features embedded in deeper layers. The LSTM encoder learns 
    a fixed length vector of time-series observed data points and the 
    LSTM decoder uses this representation to reconstruct the time-series 
    using the current hidden state and the value inferenced at the previous time-step.
    
    One interesting application example is the Encoder/Decoder for 
    Anomaly Detection (EncDec-AD) paradigm (Malhotra, P., et al. 2016).
    This reconstruction model learns to reconstruct normal time-series behavior, 
    and thereafter uses reconstruction error to detect anomalies. 
    Malhotra, P., et al. (2016) showed that EncDec-AD paradigm is robust 
    and can detect anomalies from predictable, unpredictable, periodic, aperiodic, 
    and quasi-periodic time-series. Further, they showed that the paradigm is able 
    to detect anomalies from short time-series (length as small as 30) as well as 
    long time-series (length as large as 500).

    References:
        - Cho, K., Van Merriënboer, B., Gulcehre, C., Bahdanau, D., Bougares, F., Schwenk, H., & Bengio, Y. (2014). Learning phrase representations using RNN encoder-decoder for statistical machine translation. arXiv preprint arXiv:1406.1078.
        - Malhotra, P., Ramakrishnan, A., Anand, G., Vig, L., Agarwal, P., & Shroff, G. (2016). LSTM-based encoder-decoder for multi-sensor anomaly detection. arXiv preprint arXiv:1607.00148.
    '''

    # is-a `LSTMNetworks`.
    __encoder = None

    # is-a `LSTMNetworks`.
    __decoder = None

    def __init__(
        self,
        encoder,
        decoder,
        computable_loss,
        learning_rate=1e-05,
        ctx="cpu",
        regularizatable_data_list=[],
        generating_flag=True,
        not_init_flag=False,
    ):
        '''
        Init.

        Override.

        Args:
            encoder:                        is-a `LSTMNetworks`.
            decoder:                        is-a `LSTMNetworks`.
            computable_loss:                is-a `ComputableLoss` or `mxnet.gluon.loss`.
            initializer:                    is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            batch_size:                     `int` of batch size of mini-batch.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
                                            

            seq_len:                        `int` of the length of sequences.
                                            This means refereed maxinum step `t` in feedforward.
                                            If `0`, this model will reference all series elements included 
                                            in observed data points.
                                            If not `0`, only first sequence will be observed by this model 
                                            and will be feedfowarded as feature points.
                                            This parameter enables you to build this class as `Decoder` in
                                            Sequence-to-Sequence(Seq2seq) scheme.

            hidden_n:                       `int` of the number of units in hidden layers.
            output_n:                       `int` of the nuber of units in output layer.
            dropout_rate:                   `float` of dropout rate.
            observed_activation:            `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` 
                                            that activates observed data points.

            optimizer_name:                 `str` of name of optimizer.

            input_gate_activation:          `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in input gate.
            forget_gate_activation:         `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in forget gate.
            output_gate_activation:         `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in output gate.
            hidden_activation:              `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in hidden layer.
            output_activation:              `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in output layer.
            output_layer_flag:              `bool` that means this class has output layer or not.

            ctx:                            `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            regularizatable_data_list:           `list` of `Regularizatable`.
            scale:                          `float` of scaling factor for initial parameters.
            generating_flag:                `bool` of flag. 
                                            If `False`, the `decoder` will decode all encoded data.
                                            If `True`, the `decoder`'s decoding will start at last sequence of encoded data,
                                            generating other series.
        '''
        if isinstance(encoder, LSTMNetworks) is False:
            raise TypeError("The type of `lstm_networks` must be `LSTMNetworks`.")
        if isinstance(decoder, LSTMNetworks) is False:
            raise TypeError("The type of `lstm_networks` must be `LSTMNetworks`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger
        init_deferred_flag = self.init_deferred_flag
        self.init_deferred_flag = True
        super().__init__(
            computable_loss=computable_loss,
            learning_rate=learning_rate,
            ctx=ctx,
        )
        self.init_deferred_flag = init_deferred_flag
        self.encoder = encoder
        self.decoder = decoder
        self.encoder_optimizer = None
        self.decoder_optimizer = None
        self.__computable_loss = computable_loss
        self.__generating_flag = generating_flag
        self.__not_init_flag = not_init_flag
        self.epoch = 0
        self.__input_seq_len = None
        self.__encoder_input_dim = None
        self.__decoder_input_dim = None
        self.__loss_list = []

    def parameters(self):
        '''
        '''
        return [
            {
                "params": self.encoder.parameters(),
            },
            {
                "params": self.decoder.parameters(),
            }
        ]

    def initialize_params(self, input_dim, input_seq_len):
        '''
        Initialize params.
        '''
        if self.encoder_optimizer is not None and self.decoder_optimizer is not None:
            return

        self.__input_seq_len = input_seq_len
        self.__encoder_input_dim = input_dim
        if self.init_deferred_flag is False:
            if self.__not_init_flag is False:
                self.encoder.initialize_params(
                    self.__encoder_input_dim,
                    input_seq_len
                )
                self.__decoder_input_dim = self.encoder.output_dim
                self.decoder.initialize_params(
                    self.__decoder_input_dim,
                    input_seq_len
                )
                self.encoder_optimizer = self.encoder.optimizer
                self.decoder_optimizer = self.decoder.optimizer

    def learn(self, iteratable_data):
        '''
        Learn samples drawn by `IteratableData.generate_learned_samples()`.

        Args:
            iteratable_data:     is-a `IteratableData`.
        '''
        if isinstance(iteratable_data, IteratableData) is False:
            raise TypeError("The type of `iteratable_data` must be `IteratableData`.")

        self.__loss_list = []
        try:
            epoch = self.epoch
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr in iteratable_data.generate_learned_samples():
                batch_size = batch_observed_arr.shape[0]
                seq_len = batch_observed_arr.shape[1]
                self.initialize_params(
                    input_dim=batch_observed_arr.reshape((batch_size, seq_len, -1)).shape[2],
                    input_seq_len=seq_len
                )
                self.encoder_optimizer.zero_grad()
                self.decoder_optimizer.zero_grad()
                # rank-3
                pred_arr = self.inference(batch_observed_arr)
                loss = self.compute_loss(
                    pred_arr,
                    batch_target_arr
                )
                loss.backward()
                self.encoder_optimizer.step()
                self.decoder_optimizer.step()
                self.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    with torch.inference_mode():
                        # rank-3
                        test_pred_arr = self.inference(test_batch_observed_arr)

                        test_loss = self.compute_loss(
                            test_pred_arr,
                            test_batch_target_arr
                        )
                    _loss = loss.to('cpu').detach().numpy().copy()
                    _test_loss = test_loss.to('cpu').detach().numpy().copy()
                    self.__loss_list.append((_loss, _test_loss))
                    self.__logger.debug("Epochs: " + str(epoch + 1) + " Train loss: " + str(_loss) + " Test loss: " + str(_test_loss))
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.epoch = epoch
        self.__logger.debug("end. ")

    def forward(self, x):
        '''
        Forward with torch.

        Args:
            x:      `tensor` of observed data points.
        
        Returns:
            `tensor` of inferenced feature points.
        '''
        batch_size = x.shape[0]
        seq_len = x.shape[1]
        x = torch.reshape(
            x, 
            shape=(
                batch_size, 
                seq_len, 
                -1
            )
        )
        self.initialize_params(
            input_dim=x.shape[2],
            input_seq_len=seq_len,
        )

        encoded_arr = self.encoder(x)
        self.feature_points_arr = encoded_arr

        if self.__generating_flag is True:
            arr = torch.unsqueeze(encoded_arr[:, -1, :], axis=1)
        else:
            arr = torch.flip(encoded_arr, dims=(1, ))

        decoded_arr = self.decoder(arr)
        decoded_arr = torch.flip(decoded_arr, dims=(1, ))

        self.__pred_arr = decoded_arr
        return decoded_arr

    def __rename_file(self, filename):
        filename_list = filename.split(".")
        _format = filename_list[-1]
        encoder_filename = filename.replace("." + _format, "_encoder." + _format)
        decoder_filename = filename.replace("." + _format, "_decoder." + _format)
        return encoder_filename, decoder_filename

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        encoder_filename, decoder_filename = self.__rename_file(filename)
        self.encoder.save_parameters(encoder_filename)
        self.decoder.save_parameters(decoder_filename)
        torch.save(
            {
                'epoch': self.epoch,
                'loss': self.loss_arr,
            }, 
            filename
        )

    def load_parameters(self, filename, ctx=None, strict=True):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            Context-manager that changes the selected device.
            strict:         Whether to strictly enforce that the keys in state_dict match the keys returned by this module’s state_dict() function. Default: `True`.
        '''
        encoder_filename, decoder_filename = self.__rename_file(filename)
        self.encoder.load_parameters(encoder_filename, ctx=ctx, strict=strict)
        self.decoder.load_parameters(decoder_filename, ctx=ctx, strict=strict)

        checkpoint = torch.load(filename)
        self.epoch = checkpoint['epoch']
        self.__loss_list = checkpoint['loss'].tolist()

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    __loss_list = []

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)



# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._mxnet.neural_networks import NeuralNetworks
from accelbrainbase._mxnet._exception.init_deferred_error import InitDeferredError

from mxnet.gluon.block import HybridBlock
from mxnet import gluon
from mxnet import autograd
import numpy as np
import mxnet as mx
from mxnet import MXNetError
from logging import getLogger


class AutoEncoder(NeuralNetworks):
    '''
    Auto-Encoder.

    References:
        - Kamyshanska, H., & Memisevic, R. (2014). The potential energy of an autoencoder. IEEE transactions on pattern analysis and machine intelligence, 37(6), 1261-1273.
    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self,
        encoder,
        decoder,
        computable_loss,
        initializer=None,
        learning_rate=1e-05,
        learning_attenuate_rate=1.0,
        attenuate_epoch=50,
        units_list=[100, 1],
        dropout_rate_list=[0.0, 0.5],
        optimizer_name="SGD",
        activation_list=["tanh", "sigmoid"],
        ctx=mx.gpu(),
        hybridize_flag=True,
        regularizatable_data_list=[],
        scale=1.0,
        tied_weights_flag=False,
        **kwargs
    ):
        '''
        Init.

        Args:
            encoder:                        is-a `NeuralNetworks`.
            decoder:                        is-a `NeuralNetworks`.
            computable_loss:                is-a `ComputableLoss` or `mxnet.gluon.loss`.
            initializer:                    is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
                                            

            units_list:                     `list` of int` of the number of units in hidden/output layers.
            dropout_rate_list:              `list` of `float` of dropout rate.
            observed_activation:            `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` 
                                            that activates observed data points.

            optimizer_name:                 `str` of name of optimizer.

            activation_list:                `list` of act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in input gate.

            ctx:                            `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            regularizatable_data_list:      `list` of `RegularizatableData`.
            scale:                          `float` of scaling factor for initial parameters.
            tied_weights_flag:              `bool` of flag to tied weights or not.
        '''
        if isinstance(encoder, NeuralNetworks) is False:
            raise TypeError("The type of `encoder` must be `NeuralNetworks`.")
        if isinstance(decoder, NeuralNetworks) is False:
            raise TypeError("The type of `decoder` must be `NeuralNetworks`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger
        init_deferred_flag = self.init_deferred_flag
        self.init_deferred_flag = True
        super().__init__(
            computable_loss=computable_loss,
            initializer=initializer,
            learning_rate=learning_rate,
            learning_attenuate_rate=learning_attenuate_rate,
            attenuate_epoch=attenuate_epoch,
            units_list=units_list,
            dropout_rate_list=dropout_rate_list,
            optimizer_name=optimizer_name,
            activation_list=activation_list,
            ctx=ctx,
            hybridize_flag=hybridize_flag,
            regularizatable_data_list=regularizatable_data_list,
            scale=scale,
            **kwargs
        )
        self.init_deferred_flag = init_deferred_flag
        self.encoder = encoder
        self.decoder = decoder
        self.__tied_weights_flag = tied_weights_flag
        if self.init_deferred_flag is False:
            try:
                self.collect_params().initialize(self.initializer, force_reinit=True, ctx=ctx)
                self.trainer = gluon.Trainer(self.collect_params(), optimizer_name, {"learning_rate": learning_rate})
                if hybridize_flag is True:
                    self.encoder.hybridize()
                    self.decoder.hybridize()
            except InitDeferredError:
                self.__logger.debug("The initialization should be deferred.")
        self.__computable_loss = computable_loss
        self.__units_list = units_list

    def collect_params(self, select=None):
        '''
        Overrided `collect_params` in `mxnet.gluon.HybridBlok`.
        '''
        params_dict = self.encoder.collect_params(select)
        params_dict.update(self.decoder.collect_params(select))
        return params_dict

    def inference(self, observed_arr):
        '''
        Inference the feature points.

        Args:
            observed_arr:   rank-2 Array like or sparse matrix as the observed data points.
                            The shape is: (batch size, feature points)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self(observed_arr)

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `mxnet.ndarray` or `mxnet.symbol`.
            labeled_arr:    `mxnet.ndarray` or `mxnet.symbol`.

        Returns:
            loss.
        '''
        return self.__computable_loss(pred_arr, labeled_arr)

    def extract_feature_points(self):
        '''
        Extract the activities in hidden layer and reset it, 
        considering this method will be called per one cycle in instances of time-series.

        Returns:
            The `mxnet.ndarray` of array like or sparse matrix of feature points or virtual visible observed data points.
        '''
        return self.feature_points_arr

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_dict = self.collect_params()
        
        params_arr_dict = {}
        for k in params_dict:
            params_arr_dict.setdefault(k, params_dict[k].data())

        return params_arr_dict

    def hybrid_forward(self, F, x):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        # rank-3
        return self.forward_propagation(F, x)

    def forward_propagation(self, F, x):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        encoded_arr = self.encoder.forward_propagation(F, x)
        self.feature_points_arr = encoded_arr
        decoded_arr = self.decoder.forward_propagation(F, encoded_arr)
        self.__pred_arr = decoded_arr
        return decoded_arr

    def regularize(self):
        '''
        Regularization.
        '''
        self.__tie_weights()
        super().regularize()

    def __tie_weights(self):
        if self.__tied_weights_flag is True:
            encoder_params_dict = self.encoder.extract_learned_dict()
            decoder_params_dict = self.decoder.extract_learned_dict()
            encoder_weight_keys_list = [key for key in encoder_params_dict.keys() if "weight" in key]
            decoder_weight_keys_list = [key for key in decoder_params_dict.keys() if "weight" in key]

            for i in range(len(self.encoder.units_list)):
                encoder_layer = i
                decoder_layer = len(self.encoder.units_list) - i - 1
                encoder_weight_keys, decoder_weight_keys = None, None
                for _encoder_weight_keys in encoder_weight_keys_list:
                    if "_dense" + str(encoder_layer) + "_weight" in _encoder_weight_keys:
                        encoder_weight_keys = _encoder_weight_keys
                        break

                for _decoder_weight_keys in decoder_weight_keys_list:
                    if "_dense" + str(decoder_layer) + "_weight" in _decoder_weight_keys:
                        decoder_weight_keys = _decoder_weight_keys
                        break

                if encoder_weight_keys is not None and decoder_weight_keys is not None:
                    try:
                        decoder_params_dict[decoder_weight_keys] = encoder_params_dict[encoder_weight_keys].T
                    except AssertionError:
                        raise ValueError(
                            "The shapes of weight matrixs must be equivalents in encoder layer " + str(encoder_layer) + " and decoder layer " + str(decoder_layer)
                        )

            for k, params in self.decoder.collect_params().items():
                if k in decoder_weight_keys_list:
                    params.set_data(decoder_params_dict[k])

    def __rename_file(self, filename):
        filename_list = filename.split(".")
        _format = filename_list[-1]
        encoder_filename = filename.replace("." + _format, "_encoder." + _format)
        decoder_filename = filename.replace("." + _format, "_decoder." + _format)
        return encoder_filename, decoder_filename

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        encoder_filename, decoder_filename = self.__rename_file(filename)
        self.encoder.save_parameters(encoder_filename)
        self.decoder.save_parameters(decoder_filename)

    def load_parameters(self, filename, ctx=None, allow_missing=False, ignore_extra=False):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            `mx.cpu()` or `mx.gpu()`.
            allow_missing:  `bool` of whether to silently skip loading parameters not represents in the file.
            ignore_extra:   `bool` of whether to silently ignre parameters from the file that are not present in this `Block`.
        '''
        encoder_filename, decoder_filename = self.__rename_file(filename)
        self.encoder.load_parameters(encoder_filename, ctx=ctx, allow_missing=allow_missing, ignore_extra=ignore_extra)
        self.decoder.load_parameters(decoder_filename, ctx=ctx, allow_missing=allow_missing, ignore_extra=ignore_extra)

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not.'''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not.'''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)



# -*- coding: utf-8 -*-
from mxnet.gluon.loss import Loss
from accelbrainbase.computable_loss import ComputableLoss
import numpy as np
from mxnet.gluon.loss import _reshape_like, _apply_weighting


class DRCNLoss(Loss, ComputableLoss):
    '''
    Loss function of Deep Reconstruction-Classification Networks.

    Deep Reconstruction-Classification Network(DRCN) is a convolutional network 
    that jointly learns two tasks: 
    
    1. supervised source label prediction.
    2. unsupervised target data reconstruction. 

    Ideally, a discriminative representation should model both the label and 
    the structure of the data. Based on that intuition, Ghifary, M., et al.(2016) hypothesize 
    that a domain-adaptive representation should satisfy two criteria:
    
    1. classify well the source domain labeled data.
    2. reconstruct well the target domain unlabeled data, which can be viewed as an approximate of the ideal discriminative representation.

    The encoding parameters of the DRCN are shared across both tasks, 
    while the decoding parameters are sepa-rated. The aim is that the learned label 
    prediction function can perform well onclassifying images in the target domain
    thus the data reconstruction can beviewed as an auxiliary task to support the 
    adaptation of the label prediction.

    References:
        - Ghifary, M., Kleijn, W. B., Zhang, M., Balduzzi, D., & Li, W. (2016, October). Deep reconstruction-classification networks for unsupervised domain adaptation. In European Conference on Computer Vision (pp. 597-613). Springer, Cham.
    '''

    def __init__(
        self, 
        axis=-1,
        sparse_label=False,
        rc_lambda=0.75,
        from_logits=False,
        log_softmax_flag=True,
        weight=1.0,
        classification_weight=None, 
        reconstruction_weight=None, 
        grad_clip_threshold=0.0,
        batch_axis=0, 
        **kwargs
    ):
        '''
        Init.

        Args:
            axis:                   The axis to sum over when computing softmax and entropy.
            sparse_label:           Whether label is an integer array instead of probability distribution.
            rc_lambda:              Tradeoff parameter for loss functions.
            from_logits:            Whether input is a log probability (usually from log_softmax) instead of unnormalized numbers.
            log_softmax_flag:       Use `F.log_softmax` when `from_logits` is `False`.
                                    If this value is `False`, this class will use not `F.log_softmax` but `F.softmax`.

            weight:                 Global scalar weight for total loss.
            classification_weight:  Global scalar weight for classification loss. If `None`, this value will be equivalent to `weight`.
            reconstruction_weight:  Global scalar weight for reconstruction loss. If `None`, this value will be equivalent to `weight`.
            grad_clip_threshold:    Threshold of the gradient clipping.
            batch_axis:             The axisthat represents mini-batch.
            **kwargs:               Any parmaeter.
        '''
        super(DRCNLoss, self).__init__(
            weight,
            batch_axis, 
            **kwargs
        )
        self._axis = axis
        self._sparse_label = sparse_label
        self._from_logits = from_logits
        self.__log_softmax_flag = log_softmax_flag
        self.__rc_lambda = rc_lambda
        self.__classification_weight = classification_weight
        self.__reconstruction_weight = reconstruction_weight
        self.__grad_clip_threshold = grad_clip_threshold

    def compute(
        self, 
        decoded_arr, 
        pred_arr, 
        observed_arr, 
        label_arr, 
    ):
        '''
        Compute loss.

        Args:
            decoded_arr:            `mxnet.ndarray` or `mxnet.symbol` of decoded feature points.
            pred_arr:               `mxnet.ndarray` or `mxnet.symbol` of inferenced labeled feature points.
            observed_arr:           `mxnet.ndarray` or `mxnet.symbol` of observed data points.
            label_arr:              `mxnet.ndarray` or `mxnet.symbol` of label data.

        Returns:
            Tensor of losses.
        '''
        return self(
            decoded_arr, 
            pred_arr, 
            observed_arr, 
            label_arr, 
        )

    def hybrid_forward(
        self, 
        F, 
        decoded_arr, 
        pred_arr, 
        observed_arr, 
        label_arr, 
        sample_weight=None
    ):
        '''
        Forward propagation, computing losses.

        Args:
            F:                      `mxnet.ndarray` or `mxnet.symbol`.
            decoded_arr:            `mxnet.ndarray` or `mxnet.symbol` of decoded feature points.
            pred_arr:               `mxnet.ndarray` or `mxnet.symbol` of inferenced labeled feature points.
            observed_arr:           `mxnet.ndarray` or `mxnet.symbol` of observed data points.
            label_arr:              `mxnet.ndarray` or `mxnet.symbol` of label data.
            sample_weight:          element-wise weighting tensor. 
                                    Must be broadcastable to the same shape as label. 
                                    For example, if label has shape (64, 10) and you want to weigh 
                                    each sample in the batch separately, sample_weight should have shape (64, 1).

        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of loss.
        '''
        if not self._from_logits:
            if self.__log_softmax_flag is True:
                pred_arr = F.log_softmax(pred_arr, self._axis)
            else:
                pred_arr = pred_arr - F.reshape(F.max(pred_arr, axis=self._axis), shape=(-1, 1))
                pred_arr = F.exp(pred_arr)
                pred_arr = pred_arr / F.reshape(F.sum(pred_arr, axis=self._axis), shape=(-1, 1))

        if self._sparse_label:
            classification_loss_arr = -F.pick(pred_arr, label_arr, axis=self._axis, keepdims=True)
        else:
            label_arr = _reshape_like(F, label_arr, pred_arr)
            classification_loss_arr = -F.sum(pred_arr * label_arr, axis=self._axis, keepdims=True)

        if self.__grad_clip_threshold > 0:
            classification_loss_norm = F.norm(classification_loss_arr)
            if classification_loss_norm.asscalar() > self.__grad_clip_threshold:
                classification_loss_arr = classification_loss_arr * self.__grad_clip_threshold / classification_loss_norm

        if self.__classification_weight is None:
            classification_loss_arr = _apply_weighting(F, classification_loss_arr, self._weight, sample_weight)
        else:
            classification_loss_arr = _apply_weighting(F, classification_loss_arr, self.__classification_weight, sample_weight)

        classification_loss_arr = _apply_weighting(F, classification_loss_arr, self.__rc_lambda, sample_weight)
        classification_loss = F.mean(classification_loss_arr, axis=self._batch_axis, exclude=True)

        observed_arr = _reshape_like(F, observed_arr, decoded_arr)
        reconstruction_loss_arr = F.square(observed_arr - decoded_arr)

        if self.__grad_clip_threshold > 0:
            reconstruction_loss_norm = F.norm(reconstruction_loss_arr)
            if reconstruction_loss_norm.asscalar() > self.__grad_clip_threshold:
                reconstruction_loss_arr = reconstruction_loss_arr * self.__grad_clip_threshold / reconstruction_loss_norm

        if self.__reconstruction_weight is None:
            reconstruction_loss_arr = _apply_weighting(F, reconstruction_loss_arr, self._weight / 2, sample_weight)
        else:
            reconstruction_loss_arr = _apply_weighting(F, reconstruction_loss_arr, self.__reconstruction_weight / 2, sample_weight)

        reconstruction_loss_arr = _apply_weighting(F, reconstruction_loss_arr, (1 - self.__rc_lambda), sample_weight)
        reconstruction_loss = F.mean(reconstruction_loss_arr, axis=self._batch_axis, exclude=True)

        return classification_loss + reconstruction_loss, classification_loss, reconstruction_loss



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod


class ComputableLoss(metaclass=ABCMeta):
    '''
    The interface of Loss function.
    '''

    @abstractmethod
    def compute(
        self,
        pred_arr, 
        real_arr,
    ):
        '''
        Compute loss.

        Args:
            pred_arr:       Inferenced results.
            real_arr:       Real results.
        
        Returns:
            Tensor of losses.
        '''
        raise NotImplementedError()



# -*- coding: utf-8 -*-
import numpy as np

import torch

import pandas as pd
from logging import getLogger
import os

from accelbrainbase.extractabledata.unlabeled_csv_extractor import UnlabeledCSVExtractor
from accelbrainbase.iteratabledata.labeled_image_iterator import LabeledImageIterator as _LabeledImageIterator
from accelbrainbase.noiseable_data import NoiseableData


class UnlabeledCSVIterator(_LabeledImageIterator):
    '''
    Iterator that draws from CSV files and generates `mxnet.ndarray` of unlabeled samples.
    '''

    def __init__(
        self,
        unlabeled_csv_extractor,
        train_csv_path,
        test_csv_path,
        epochs=300,
        batch_size=20,
        norm_mode="z_score",
        scale=1.0,
        noiseable_data=None,
        ctx="cpu"
    ):
        '''
        Init.

        Args:
            image_extractor:                is-a `ImageExtractor`.
            train_csv_path:                 `str` of path to CSV file in training.
            test_csv_path:                  `str` of path to CSV file in test.

            epochs:                         `int` of epochs of Mini-batch.
            bath_size:                      `int` of batch size of Mini-batch.
            norm_mode:                      How to normalize pixel values of images.
                                            - `z_score`: Z-Score normalization.
                                            - `min_max`: Min-max normalization.
                                            - others : This class will not normalize the data.

            scale:                          `float` of scaling factor for data.
            noiseable_data:                 is-a `NoiseableData` for Denoising Auto-Encoders.
        '''
        if isinstance(unlabeled_csv_extractor, UnlabeledCSVExtractor) is False:
            raise TypeError("The type of `unlabeled_csv_extractor` must be `UnlabeledCSVExtractor`.")
        if noiseable_data is not None and isinstance(noiseable_data, NoiseableData) is False:
            raise TypeError("The type of `noiseable_data` must be `NoiseableData`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        train_observed_arr = unlabeled_csv_extractor.extract(train_csv_path)
        test_observed_arr = unlabeled_csv_extractor.extract(test_csv_path)

        dataset_size = train_observed_arr.shape[0]
        iter_n = int(epochs * max(dataset_size / batch_size, 1))

        self.__train_observed_arr = train_observed_arr
        self.__test_observed_arr = test_observed_arr

        self.iter_n = iter_n
        self.epochs = epochs
        self.batch_size = batch_size
        self.norm_mode = norm_mode
        self.scale = scale
        self.__noiseable_data = noiseable_data

        self.__ctx = ctx

    def generate_learned_samples(self):
        '''
        Draw and generate data.

        Returns:
            `Tuple` data. The shape is ...
            - `mxnet.ndarray` of observed data points in training.
            - `mxnet.ndarray` of supervised data in training.
            - `mxnet.ndarray` of observed data points in test.
            - `mxnet.ndarray` of supervised data in test.
        '''
        for _ in range(self.iter_n):
            training_batch_arr, test_batch_arr = None, None

            row_arr = np.arange(self.__train_observed_arr.shape[0])
            np.random.shuffle(row_arr)

            training_batch_arr = self.__train_observed_arr[row_arr[:self.batch_size]]
            training_batch_arr = torch.from_numpy(training_batch_arr)
            training_batch_arr = training_batch_arr.to(self.__ctx).float()
            training_batch_arr = self.pre_normalize(training_batch_arr)

            test_row_arr = np.arange(self.__test_observed_arr.shape[0])
            np.random.shuffle(test_row_arr)

            test_batch_arr = self.__test_observed_arr[test_row_arr[:self.batch_size]]
            test_batch_arr = torch.from_numpy(test_batch_arr)
            test_batch_arr = test_batch_arr.to(self.__ctx).float()
            test_batch_arr = self.pre_normalize(test_batch_arr)

            if self.__noiseable_data is not None:
                training_batch_arr = self.__noiseable_data.noise(training_batch_arr)

            yield training_batch_arr, training_batch_arr, test_batch_arr, test_batch_arr

    def generate_inferenced_samples(self):
        '''
        Draw and generate data.
        The targets will be drawn from all image file sorted in ascending order by file name.

        Returns:
            `Tuple` data. The shape is ...
            - `None`.
            - `None`.
            - `mxnet.ndarray` of observed data points in test.
            - file path.
        '''
        for i in range(1, self.__test_observed_arr.shape[0] // self.batch_size):
            test_batch_arr = self.__test_observed_arr[(i-1)*self.batch_size:i*self.batch_size]
            test_batch_arr = torch.from_numpy(test_batch_arr)
            test_batch_arr = test_batch_arr.to(self.__ctx).float()
            yield None, None, test_batch_arr, None

    def pre_normalize(self, arr):
        '''
        Normalize before observation.

        Args:
            arr:    Tensor.
        
        Returns:
            Tensor.
        '''
        if self.norm_mode == "min_max":
            if torch.max(arr) != torch.min(arr):
                n = 0.0
            else:
                n = 1e-08
            arr = (arr - torch.min(arr)) / (torch.max(arr) - torch.min(arr) + n)
        elif self.norm_mode == "z_score":
            std = torch.std(arr)
            if std == 0:
                std += 1e-08
            arr = (arr - torch.mean(arr)) / std

        arr = arr * self.scale
        return arr



# -*- coding: utf-8 -*-
from accelbrainbase.iteratabledata._mxnet.drcn_iterator import DRCNIterator
from accelbrainbase.regularizatable_data import RegularizatableData
from accelbrainbase.observabledata._mxnet.neural_networks import NeuralNetworks
from accelbrainbase.computableloss._mxnet.drcn_loss import DRCNLoss
from accelbrainbase.observabledata._mxnet.convolutional_neural_networks import ConvolutionalNeuralNetworks
from accelbrainbase.observabledata._mxnet.convolutionalneuralnetworks.convolutional_auto_encoder import ConvolutionalAutoEncoder
from accelbrainbase._mxnet._exception.init_deferred_error import InitDeferredError

from mxnet import gluon
from mxnet import autograd
import numpy as np
import mxnet as mx
from mxnet import MXNetError
from logging import getLogger


class DRCNetworks(ConvolutionalNeuralNetworks):
    '''
    Deep Reconstruction-Classification Networks(DRCN or DRCNetworks).

    Deep Reconstruction-Classification Network(DRCN or DRCNetworks) is a convolutional network 
    that jointly learns two tasks: 
    
    1. supervised source label prediction.
    2. unsupervised target data reconstruction. 

    Ideally, a discriminative representation should model both the label and 
    the structure of the data. Based on that intuition, Ghifary, M., et al.(2016) hypothesize 
    that a domain-adaptive representation should satisfy two criteria:
    
    1. classify well the source domain labeled data.
    2. reconstruct well the target domain unlabeled data, which can be viewed as an approximate of the ideal discriminative representation.

    The encoding parameters of the DRCN are shared across both tasks, 
    while the decoding parameters are sepa-rated. The aim is that the learned label 
    prediction function can perform well onclassifying images in the target domain
    thus the data reconstruction can beviewed as an auxiliary task to support the 
    adaptation of the label prediction.

    References:
        - Ghifary, M., Kleijn, W. B., Zhang, M., Balduzzi, D., & Li, W. (2016, October). Deep reconstruction-classification networks for unsupervised domain adaptation. In European Conference on Computer Vision (pp. 597-613). Springer, Cham.
    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    # `list` of losses.
    __loss_list = []

    # `list` of accuracies.
    __acc_list = []

    def __init__(
        self,
        convolutional_auto_encoder,
        drcn_loss,
        initializer=None,
        learning_rate=1e-05,
        learning_attenuate_rate=1.0,
        attenuate_epoch=50,
        hidden_units_list=[],
        output_nn=None,
        hidden_dropout_rate_list=[],
        optimizer_name="SGD",
        hidden_activation_list=[],
        hidden_batch_norm_list=[],
        ctx=mx.gpu(),
        hybridize_flag=True,
        regularizatable_data_list=[],
        scale=1.0,
        tied_weights_flag=True,
        tol=3.0,
        est=1e-08,
        wd=0.0,
        **kwargs
    ):
        '''
        Init.

        Args:
            convolutional_auto_encoder:     is-a `ConvolutionalAutoEncoder`.
            drcn_loss:                      is-a `DRCNLoss`.
            initializer:                    is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
                                            

            hidden_units_list:              `list` of `mxnet.gluon.nn._conv` in hidden layers.
            output_nn:                      is-a `NeuralNetworks` as output layers.
                                            If `None`, last layer in `hidden_units_list` will be considered as an output layer.

            hidden_dropout_rate_list:       `list` of `float` of dropout rate in hidden layers.

            optimizer_name:                 `str` of name of optimizer.

            hidden_activation_list:         `list` of act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in input gate.
            hidden_batch_norm_list:         `list` of `mxnet.gluon.nn.BatchNorm` in hidden layers.

            ctx:                            `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:                  `bool` of flag that means this class will call `mxnet.gluon.HybridBlock.hybridize()` or not.
            regularizatable_data_list:           `list` of `Regularizatable`.
            scale:                          `float` of scaling factor for initial parameters.
            tied_weights_flag:              `bool` of flag to tied weights or not.
            tol:                            `float` of the skipping epoch.
                                            If it is not `None`, the epoch will be skipped and parameters will be not updated 
                                            when (loss > previous_loss - tol).

            est:                            `float` of the early stopping.
                                            If it is not `None`, the learning will be stopped 
                                            when (|loss -previous_loss| < est).

        '''
        if isinstance(convolutional_auto_encoder, ConvolutionalAutoEncoder) is False:
            raise TypeError("The type of `convolutional_auto_encoder` must be `ConvolutionalAutoEncoder`.")

        if len(hidden_units_list) != len(hidden_activation_list):
            raise ValueError("The length of `hidden_units_list` and `hidden_activation_list` must be equivalent.")

        if len(hidden_dropout_rate_list) != len(hidden_units_list):
            raise ValueError("The length of `hidden_dropout_rate_list` and `hidden_units_list` must be equivalent.")

        if isinstance(drcn_loss, DRCNLoss) is False:
            raise TypeError("The type of `drcn_loss` must be `DRCNLoss`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger
        init_deferred_flag = self.init_deferred_flag
        self.init_deferred_flag = True

        super().__init__(
            computable_loss=drcn_loss,
            initializer=initializer,
            learning_rate=learning_rate,
            learning_attenuate_rate=learning_attenuate_rate,
            attenuate_epoch=attenuate_epoch,
            hidden_units_list=hidden_units_list,
            output_nn=None,
            hidden_dropout_rate_list=hidden_dropout_rate_list,
            optimizer_name=optimizer_name,
            hidden_activation_list=hidden_activation_list,
            hidden_batch_norm_list=hidden_batch_norm_list,
            ctx=ctx,
            hybridize_flag=hybridize_flag,
            regularizatable_data_list=regularizatable_data_list,
            scale=scale,
            **kwargs
        )
        self.init_deferred_flag = init_deferred_flag
        self.convolutional_auto_encoder = convolutional_auto_encoder
        self.__tied_weights_flag = tied_weights_flag

        self.output_nn = None

        if initializer is None:
            self.initializer = mx.initializer.Xavier(
                rnd_type="gaussian", 
                factor_type="in", 
                magnitude=1
            )
        else:
            if isinstance(initializer, mx.initializer.Initializer) is False:
                raise TypeError("The type of `initializer` must be `mxnet.initializer.Initializer`.")

            self.initializer = initializer

        if self.init_deferred_flag is False:
            try:
                self.collect_params().initialize(
                    self.initializer,
                    force_reinit=True, 
                    ctx=ctx
                )
                self.trainer = gluon.Trainer(
                    self.collect_params(), 
                    optimizer_name, 
                    {
                        "learning_rate": learning_rate,
                        "wd": wd,
                    }
                )

                if output_nn is not None:
                    output_nn.collect_params().initialize(
                        output_nn.initializer,
                        force_reinit=True,
                        ctx=ctx
                    )
                    self.output_nn_trainer = gluon.Trainer(
                        output_nn.collect_params(),
                        optimizer_name,
                        {
                            "learning_rate": learning_rate,
                            "wd": wd,
                        }
                    )
                else:
                    self.output_nn_trainer = None

                self.output_nn = output_nn

                if hybridize_flag is True:
                    self.convolutional_auto_encoder.encoder.hybridize()
                    self.convolutional_auto_encoder.decoder.hybridize()

                    if self.output_nn is not None:
                        self.output_nn.hybridize()

            except InitDeferredError:
                self.__logger.debug("The initialization should be deferred.")

        for v in regularizatable_data_list:
            if isinstance(v, RegularizatableData) is False:
                raise TypeError("The type of values of `regularizatable_data_list` must be `Regularizatable`.")
        self.__regularizatable_data_list = regularizatable_data_list

        self.drcn_loss = drcn_loss

        self.__learning_rate = learning_rate
        self.__learning_attenuate_rate = learning_attenuate_rate
        self.__attenuate_epoch = attenuate_epoch

        self.__ctx = ctx

        self.__loss_list = []
        self.__acc_list = []
        self.__target_domain_arr = None

        self.__tied_weights_flag = tied_weights_flag

        self.__tol = tol
        self.__est = est

    def collect_params(self, select=None):
        '''
        Overrided `collect_params` in `mxnet.gluon.HybridBlok`.
        '''
        params_dict = super().collect_params(select)
        params_dict.update(self.convolutional_auto_encoder.collect_params(select))

        if self.output_nn is not None:
            params_dict.update(self.output_nn.collect_params(select))

        return params_dict

    def learn(self, iteratable_data):
        '''
        Learn the observed data points with domain adaptation.

        Args:
            iteratable_data:     is-a `DRCNIterator`.

        '''
        if isinstance(iteratable_data, DRCNIterator) is False:
            raise TypeError("The type of `iteratable_data` must be `DRCNIterator`.")

        self.__loss_list = []
        self.__acc_list = []
        learning_rate = self.__learning_rate
        self.__previous_loss = None
        tol_flag = False
        est_flag = False
        try:
            epoch = 0
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr, target_domain_arr in iteratable_data.generate_learned_samples():
                self.epoch = epoch
                self.batch_size = batch_observed_arr.shape[0]
                self.convolutional_auto_encoder.batch_size = self.batch_size
                if ((epoch + 1) % self.__attenuate_epoch == 0):
                    learning_rate = learning_rate * self.__learning_attenuate_rate
                    self.trainer.set_learning_rate(learning_rate)
                    if self.output_nn_trainer is not None:
                        self.output_nn_trainer.set_learning_rate(learning_rate)

                with autograd.record():
                    # rank-3
                    decoded_arr = self.inference_auto_encoder(target_domain_arr)
                    _, prob_arr = self.inference(batch_observed_arr)
                    loss, train_classification_loss, train_reconstruction_loss = self.compute_loss(
                        decoded_arr, 
                        prob_arr,
                        target_domain_arr,
                        batch_target_arr
                    )

                loss_mean = loss.mean().asscalar()
                if self.__previous_loss is not None:
                    loss_diff = loss_mean - self.__previous_loss
                    if self.__tol is not None and loss_diff > self.__tol:
                        tol_flag = True
                    if self.__est is not None and np.abs(loss_diff) < self.__est:
                        est_flag = True

                if est_flag is False:
                    if tol_flag is True:
                        if ((epoch + 1) % 100 == 0):
                            self.__logger.debug("TOL...")
                            self.__logger.debug(str(self.__previous_loss) + " -> " + str(loss_mean))
                        self.trainer.set_learning_rate(0.0)
                        if self.output_nn_trainer is not None:
                            self.output_nn_trainer.set_learning_rate(0.0)

                    loss.backward()
                    self.trainer.step(batch_observed_arr.shape[0])
                    if self.output_nn_trainer is not None:
                        self.output_nn_trainer.step(batch_observed_arr.shape[0])
                    self.regularize()

                    self.__previous_loss = loss_mean

                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        # rank-3
                        test_decoded_arr = self.inference_auto_encoder(target_domain_arr)
                        _, test_prob_arr = self.inference(test_batch_observed_arr)
                        test_loss, test_classification_loss, test_reconstruction_loss = self.compute_loss(
                            test_decoded_arr, 
                            test_prob_arr,
                            target_domain_arr,
                            test_batch_target_arr
                        )

                        self.__logger.debug("Epochs: " + str(epoch + 1) + " Train total loss: " + str(loss.asnumpy().mean()) + " Test total loss: " + str(test_loss.asnumpy().mean()))
                        self.__logger.debug("Train classification loss: " + str(train_classification_loss.asnumpy().mean()) + " Test classification loss: " + str(test_classification_loss.asnumpy().mean()))
                        self.__logger.debug("Train reconstruction loss: " + str(train_reconstruction_loss.asnumpy().mean()) + " Test reconstruction loss: " + str(test_reconstruction_loss.asnumpy().mean()))

                    if self.compute_acc_flag is True:
                        if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                            acc, inferenced_label_arr, answer_label_arr = self.compute_acc(prob_arr, batch_target_arr)
                            test_acc, test_inferenced_label_arr, test_answer_label_arr = self.compute_acc(test_prob_arr, test_batch_target_arr)
                            if (epoch + 1) % 100 == 0 or epoch < 100:
                                acc, inferenced_label_arr, answer_label_arr = self.compute_acc(prob_arr, batch_target_arr)
                                test_acc, test_inferenced_label_arr, test_answer_label_arr = self.compute_acc(test_prob_arr, test_batch_target_arr)

                                self.__logger.debug("-" * 100)
                                self.__logger.debug("Train accuracy: " + str(acc) + " Test accuracy: " + str(test_acc))
                                self.__logger.debug("Train infenreced label(inferenced):")
                                self.__logger.debug(inferenced_label_arr.asnumpy())
                                self.__logger.debug("Train infenreced label(answer):")
                                self.__logger.debug(answer_label_arr.asnumpy())

                                self.__logger.debug("Test infenreced label(inferenced):")
                                self.__logger.debug(test_inferenced_label_arr.asnumpy())
                                self.__logger.debug("Test infenreced label(answer):")
                                self.__logger.debug(test_answer_label_arr.asnumpy())
                                self.__logger.debug("-" * 100)

                                if (test_answer_label_arr[0].asnumpy() == test_answer_label_arr.asnumpy()).astype(int).sum() != test_answer_label_arr.shape[0]:
                                    if (test_inferenced_label_arr[0].asnumpy() == test_inferenced_label_arr.asnumpy()).astype(int).sum() == test_inferenced_label_arr.shape[0]:
                                        self.__logger.debug("It may be overfitting.")

                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        self.__loss_list.append(
                            (
                                loss.asnumpy().mean(), 
                                test_loss.asnumpy().mean(),
                                train_classification_loss.asnumpy().mean(),
                                test_classification_loss.asnumpy().mean(),
                                train_reconstruction_loss.asnumpy().mean(),
                                test_reconstruction_loss.asnumpy().mean()
                            )
                        )
                        if self.compute_acc_flag is True:
                            self.__acc_list.append(
                                (
                                    acc,
                                    test_acc
                                )
                            )

                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        epoch += 1
                    iter_n += 1

                    if tol_flag is True:
                        self.trainer.set_learning_rate(learning_rate)
                        if self.output_nn_trainer is not None:
                            self.output_nn_trainer.set_learning_rate(learning_rate)

                else:
                    self.__logger.debug("Early stopping.")
                    self.__logger.debug(str(self.__previous_loss) + " -> " + str(loss_mean))
                    break

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.__logger.debug("end. ")

    def forward_propagation(self, F, x):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            Tuple data.
                - `mxnet.ndarray` or `mxnet.symbol` of reconstrcted feature points.
                - `mxnet.ndarray` or `mxnet.symbol` of inferenced label.
        '''
        decoded_arr = self.convolutional_auto_encoder.forward_propagation(F, x)
        self.feature_points_arr = self.convolutional_auto_encoder.feature_points_arr

        if self.output_nn is not None:
            prob_arr = self.output_nn.forward_propagation(F, self.feature_points_arr)
        else:
            prob_arr = self.feature_points_arr

        return decoded_arr, prob_arr

    def inference_auto_encoder(self, x):
        '''
        Hybrid forward with Gluon API (Auto-Encoder only).

        Args:
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of reconstrcted feature points.
        '''
        return self.convolutional_auto_encoder.inference(x)

    def regularize(self):
        '''
        Regularization.
        '''
        if self.__tied_weights_flag is True:
            self.convolutional_auto_encoder.tie_weights()
        super().regularize()

    def compute_loss(
        self, 
        decoded_arr, 
        prob_arr,
        batch_observed_arr,
        batch_target_arr
    ):
        '''
        Compute loss.

        Args:
            decoded_arr:            `mxnet.ndarray` or `mxnet.symbol` of decoded feature points..
            prob_arr:               `mxnet.ndarray` or `mxnet.symbol` of predicted labels data.
            batch_observed_arr:     `mxnet.ndarray` or `mxnet.symbol` of observed data points.
            batch_target_arr:       `mxnet.ndarray` or `mxnet.symbol` of label data.

        Returns:
            loss.
        '''
        return self.drcn_loss(
            decoded_arr, 
            prob_arr,
            batch_observed_arr,
            batch_target_arr
        )

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)

    def get_acc_list(self):
        ''' getter for accuracies. '''
        return np.array(self.__acc_list)
    
    acc_arr = property(get_acc_list, set_readonly)



# -*- coding: utf-8 -*-
from accelbrainbase.extractabledata.image_extractor import ImageExtractor as _ImageExtractor
from PIL import Image
import numpy as np
import torch


class ImageExtractor(_ImageExtractor):
    '''
    Image Extractor.
    '''

    def __init__(
        self, 
        width,
        height,
        channel,
        ctx="cpu"
    ):
        '''
        Init.

        Args:
            width:          `int` of image width.
            height:         `int` of image height.
            channel:        `int` of channel of image.
            ctx:            `mx.cpu()` or `mx.gpu()`.
        '''
        self.width = width
        self.height = height
        self.channel = channel
        self.__ctx = ctx

    def extract(
        self,
        path,
    ):
        '''
        Extract image file data.

        Args:
            path:     `str` of image files.
        
        Returns:
            `mxnet.ndarray` of data.
            The shape is (`channel`, `width`, `height`).
        '''
        img_arr = super().extract(path=path)
        img_arr = torch.from_numpy(img_arr.astype(np.float32)).clone()
        img_arr = img_arr.to(self.__ctx).float()

        return img_arr



# -*- coding: utf-8 -*-
from accelbrainbase.observable_data import ObservableData
from accelbrainbase.iteratable_data import IteratableData
from accelbrainbase.regularizatable_data import RegularizatableData
from accelbrainbase.computable_loss import ComputableLoss
from mxnet.gluon.block import HybridBlock
from mxnet import gluon
from mxnet import autograd
import numpy as np
import mxnet as mx
import mxnet.ndarray as nd
from mxnet import MXNetError
from logging import getLogger


class AttentionModel(HybridBlock, ObservableData):
    '''
    Attention Model.

    References:
        - Bahdanau, D., Cho, K., & Bengio, Y. (2014). Neural machine translation by jointly learning to align and translate. arXiv preprint arXiv:1409.0473.
        - Floridi, L., & Chiriatti, M. (2020). GPT-3: Its nature, scope, limits, and consequences. Minds and Machines, 30(4), 681-694.
        - Miller, A., Fisch, A., Dodge, J., Karimi, A. H., Bordes, A., & Weston, J. (2016). Key-value memory networks for directly reading documents. arXiv preprint arXiv:1606.03126.
        - Radford, A., Narasimhan, K., Salimans, T., & Sutskever, I. (2018) Improving Language Understanding by Generative Pre-Training. OpenAI (URL: https://s3-us-west-2.amazonaws.com/openai-assets/research-covers/language-unsupervised/language_understanding_paper.pdf)
        - Radford, A., Wu, J., Child, R., Luan, D., Amodei, D., & Sutskever, I. (2019). Language models are unsupervised multitask learners. OpenAI blog, 1(8), 9.
        - Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., & Polosukhin, I. (2017). Attention is all you need. arXiv preprint arXiv:1706.03762.

    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    __depth_dim = None

    def get_depth_dim(self):
        ''' getter '''
        return self.__depth_dim
    
    def set_depth_dim(self, value):
        ''' setter '''
        self.__depth_dim = value

    depth_dim = property(get_depth_dim, set_depth_dim)

    def __init__(
        self,
        depth_dim,
        computable_loss,
        dropout_rate=0.5,
        initializer=None,
        learning_rate=1e-05,
        learning_attenuate_rate=1.0,
        attenuate_epoch=50,
        optimizer_name="SGD",
        ctx=mx.gpu(),
        hybridize_flag=True,
        regularizatable_data_list=[],
        not_init_flag=False,
        **kwargs
    ):
        '''
        Init.

        Args:
            depth_dim:                      `int` of dimension of dense layer.
            computable_loss:                is-a `ComputableLoss` or `gluon.loss.Loss`.
            dropout_rate:                   `float` of dropout rate.
            initializer:                    is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.

            optimizer_name:                 `str` of name of optimizer.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            regularizatable_data_list:           `list` of `RegularizatableData`.
            not_init_flag:                  `bool` of whether initialize parameters or not.
        '''
        if isinstance(depth_dim, int) is False:
            raise TypeError("The type of `depth_dim` must be `int`.")
        if isinstance(computable_loss, ComputableLoss) is False and isinstance(computable_loss, gluon.loss.Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `gluon.loss.Loss`.")

        super(AttentionModel, self).__init__()
        self.__computable_loss = computable_loss

        if initializer is None:
            self.initializer = mx.initializer.Normal(1.0)
        else:
            if isinstance(initializer, mx.initializer.Initializer) is False:
                raise TypeError("The type of `initializer` must be `mxnet.initializer.Initializer`.")

            self.initializer = initializer

        with self.name_scope():
            self.query_dense_layer = gluon.nn.Dense(
                depth_dim,
                use_bias=False,
                flatten=False,
            )
            self.key_dense_layer = gluon.nn.Dense(
                depth_dim,
                use_bias=False,
                flatten=False,
            )
            self.value_dense_layer = gluon.nn.Dense(
                depth_dim,
                use_bias=False,
                flatten=False,
            )
            self.dropout = gluon.nn.Dropout(rate=dropout_rate)

            self.output_dense_layer = gluon.nn.Dense(
                depth_dim,
                use_bias=False,
                flatten=False,
            )
            self.register_child(self.query_dense_layer)
            self.register_child(self.key_dense_layer)
            self.register_child(self.value_dense_layer)
            self.register_child(self.output_dense_layer)
            self.register_child(self.dropout)

        if self.init_deferred_flag is False:
            if not_init_flag is False:
                self.collect_params().initialize(self.initializer, force_reinit=True, ctx=ctx)
                self.trainer = gluon.Trainer(self.collect_params(), optimizer_name, {"learning_rate": learning_rate})
                if hybridize_flag is True:
                    self.hybridize()

        for v in regularizatable_data_list:
            if isinstance(v, RegularizatableData) is False:
                raise TypeError("The type of values of `regularizatable_data_list` must be `RegularizatableData`.")
        self.__regularizatable_data_list = regularizatable_data_list

        self.__ctx = ctx
        self.ctx = ctx
        self.__learning_rate = learning_rate
        self.__learning_attenuate_rate = learning_attenuate_rate
        self.__attenuate_epoch = attenuate_epoch
        self.__optimizer_name = optimizer_name

        self.depth_dim = depth_dim

        logger = getLogger("accelbrainbase")
        self.__logger = logger

    def learn(self, iteratable_data):
        '''
        Learn samples drawn by `IteratableData.generate_learned_samples()`.

        Args:
            iteratable_data:     is-a `IteratableData`.
        '''
        if isinstance(iteratable_data, IteratableData) is False:
            raise TypeError("The type of `iteratable_data` must be `IteratableData`.")

        self.__loss_list = []
        learning_rate = self.__learning_rate

        pre_batch_observed_arr = None
        pre_test_batch_observed_arr = None
        try:
            epoch = 0
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr in iteratable_data.generate_learned_samples():
                self.epoch = epoch
                if ((epoch + 1) % self.__attenuate_epoch == 0):
                    learning_rate = learning_rate * self.__learning_attenuate_rate
                    self.trainer.set_learning_rate(learning_rate)

                with autograd.record():
                    # Self-Attention.
                    if batch_observed_arr.ndim == 2:
                        batch_observed_arr = nd.expand_dims(batch_observed_arr, axis=1)

                    pred_arr = self.inference(batch_observed_arr, batch_observed_arr)
                    loss = self.compute_loss(
                        pred_arr,
                        batch_observed_arr
                    )
                loss.backward()
                self.trainer.step(batch_observed_arr.shape[0])
                self.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    if test_batch_observed_arr.ndim == 2:
                        test_batch_observed_arr = nd.expand_dims(test_batch_observed_arr, axis=1)

                    test_pred_arr = self.inference(test_batch_observed_arr, test_batch_observed_arr)

                    test_loss = self.compute_loss(
                        test_pred_arr,
                        test_batch_observed_arr
                    )
                    self.__loss_list.append((loss.asnumpy().mean(), test_loss.asnumpy().mean()))
                    self.__logger.debug("Epochs: " + str(epoch + 1) + " Train loss: " + str(loss.asnumpy().mean()) + " Test loss: " + str(test_loss.asnumpy().mean()))
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.__logger.debug("end. ")

    def inference(self, observed_arr, memory_arr):
        '''
        Inference samples drawn by `IteratableData.generate_inferenced_samples()`.

        Args:
            observed_arr:   Array like or sparse matrix as the observed data points.
            memory_arr:     Array like or sparse matrix as the observed data points.

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self(observed_arr, memory_arr)

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `mxnet.ndarray` or `mxnet.symbol`.
            labeled_arr:    `mxnet.ndarray` or `mxnet.symbol`.

        Returns:
            loss.
        '''
        return self.__computable_loss(pred_arr, labeled_arr)

    def regularize(self):
        '''
        Regularization.
        '''
        params_dict = self.extract_learned_dict()
        for regularizatable in self.__regularizatable_data_list:
            params_dict = regularizatable.regularize(params_dict)

        for k, params in self.collect_params().items():
            params.set_data(params_dict[k])

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_dict = self.collect_params()
        
        params_arr_dict = {}
        for k in params_dict:
            params_arr_dict.setdefault(k, params_dict[k].data())

        return params_arr_dict

    def hybrid_forward(self, F, x, m):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of observed data points.
            m:      `mxnet.ndarray` of observed data points. The shape is (batch_size, length of memory, depth).
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        # rank-3
        return self.forward_propagation(F, x, m)

    def forward_propagation(self, F, x, m):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of query. The shape is (batch_size, length of query, depth).
            m:      `mxnet.ndarray` of memory. The shape is (batch_size, length of memory, depth).
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        query_arr = self.query_dense_layer(x)
        key_arr = self.key_dense_layer(m)
        value_arr = self.value_dense_layer(m)
        logit_arr = F.batch_dot(query_arr, key_arr, transpose_b=True)
        attention_weight_arr = F.softmax(logit_arr)
        attention_weight_arr = self.dropout(attention_weight_arr)
        attention_output_arr = F.batch_dot(attention_weight_arr, value_arr)
        output_arr = self.output_dense_layer(attention_output_arr)
        return output_arr

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not. '''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not. '''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)

    def get_units_list(self):
        ''' getter for `list` of units in each layer. '''
        return self.__units_list
    
    units_list = property(get_units_list, set_readonly)

    # is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
    __initializer = None

    def get_initializer(self):
        ''' getter for `mxnet.initializer`. '''
        return self.__initializer
    
    def set_initializer(self, value):
        ''' setter for `mxnet.initializer`.'''
        self.__initializer = value
    
    initializer = property(get_initializer, set_initializer)



# -*- coding: utf-8 -*-
import torch
import torch.nn as nn
from logging import getLogger
import os

from accelbrainbase.iteratabledata.labeled_image_iterator import LabeledImageIterator as _LabeledImageIterator
from accelbrainbase.noiseable_data import NoiseableData


class GaussIterator(_LabeledImageIterator):
    '''
    Iterator that draws from CSV files and generates `mxnet.ndarray`.
    '''

    def __init__(
        self,
        loc=0.0,
        std=1.0,
        dim=100,
        epochs=300,
        batch_size=20,
        norm_mode="z_score",
        noiseable_data=None,
        scale=1.0,
        ctx="cpu",
        dataset_size=1000
    ):
        '''
        Init.

        Args:
            loc:                            Mean (“centre”) of the distribution.
            std:                            Scale of the distribution.
            dim:                            Dimension.

            epochs:                         `int` of epochs of Mini-batch.
            batch_size:                      `int` of batch size of Mini-batch.
            norm_mode:                      How to normalize pixel values of images.
                                            - `z_score`: Z-Score normalization.
                                            - `min_max`: Min-max normalization.
                                            - others : This class will not normalize the data.

            noiseable_data:                 is-a `NoiseableData` for Denoising Auto-Encoders.
            dataset_size:                   `int` of the dataset size.
        '''
        if noiseable_data is not None and isinstance(noiseable_data, NoiseableData) is False:
            raise TypeError("The type of `noiseable_data` must be `NoiseableData`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        iter_n = int(epochs * max(dataset_size / batch_size, 1))

        self.__loc = loc
        self.__std = std
        self.__dim = dim

        self.iter_n = iter_n
        self.epochs = epochs
        self.batch_size = batch_size
        self.norm_mode = norm_mode
        self.scale = scale
        self.__noiseable_data = noiseable_data

        self.__ctx = ctx

    def generate_learned_samples(self):
        '''
        Draw and generate data.

        Returns:
            `Tuple` data. The shape is ...
            - `mxnet.ndarray` of observed data points in training.
            - `mxnet.ndarray` of supervised data in training.
            - `mxnet.ndarray` of observed data points in test.
            - `mxnet.ndarray` of supervised data in test.
        '''
        for _ in range(self.iter_n):
            training_batch_arr, test_batch_arr = None, None

            training_batch_arr = torch.normal(
                mean=self.__loc,
                std=self.__std,
                size=(
                    self.batch_size,
                    self.__dim
                ),
            )
            test_batch_arr = torch.normal(
                mean=self.__loc,
                std=self.__std,
                size=(
                    self.batch_size,
                    self.__dim
                ),
            )

            training_batch_arr = training_batch_arr.to(self.__ctx).float()
            test_batch_arr = test_batch_arr.to(self.__ctx).float()

            training_batch_arr = self.pre_normalize(training_batch_arr)
            test_batch_arr = self.pre_normalize(test_batch_arr)

            if self.__noiseable_data is not None:
                training_batch_arr = self.__noiseable_data.noise(training_batch_arr)

            yield training_batch_arr, training_batch_arr, test_batch_arr, test_batch_arr

    def pre_normalize(self, arr):
        '''
        Normalize before observation.

        Args:
            arr:    Tensor.
        
        Returns:
            Tensor.
        '''
        if self.norm_mode == "min_max":
            if torch.max(arr) != torch.min(arr):
                n = 0.0
            else:
                n = 1e-08
            arr = (arr - torch.min(arr)) / (torch.max(arr) - torch.min(arr) + n)
        elif self.norm_mode == "z_score":
            std = torch.std(arr)
            if std == 0:
                std += 1e-08
            arr = (arr - torch.mean(arr)) / std

        arr = arr * self.scale
        return arr

    def generate_inferenced_samples(self):
        '''
        Draw and generate data.
        The targets will be drawn from all image file sorted in ascending order by file name.

        Returns:
            `Tuple` data. The shape is ...
            - `None`.
            - `None`.
            - `mxnet.ndarray` of observed data points in test.
            - file path.
        '''
        raise NotImplementedError()
